# coding: utf-8
"""
Implementation of validate_schema(schema)
"""
import copy
import datetime
import math
import string
import typing
import urllib.parse
import re

import pint

from ..errors import ValidationError, ActionDoesNotExistError, InvalidNumberError, InvalidTemplateIDError, RecursiveTemplateError
from .utils import units_are_valid
from .validate import validate
from .templates import substitute_templates
from . import calculations
from .conditions import validate_condition_schema
from ..languages import get_language_codes
from .. import datatypes

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'


def validate_schema(
        schema: typing.Dict[str, typing.Any],
        path: typing.Optional[typing.List[str]] = None,
        *,
        root_schema: typing.Optional[typing.Dict[str, typing.Any]] = None,
        parent_conditions: typing.Optional[typing.List[typing.Tuple[typing.List[str], typing.Dict[str, typing.Any]]]] = None,
        invalid_template_action_ids: typing.Sequence[int] = (),
        strict: bool = False,
        all_language_codes: typing.Optional[typing.Set[str]] = None
) -> None:
    """
    Validates the given schema and raises a ValidationError if it is invalid.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :param root_schema: the root object schema
    :param parent_conditions: conditions defined in parent objects
    :param invalid_template_action_ids: IDs of actions that may not be used as templates to prevent recursion
    :param strict: whether the schema should be evaluated in strict mode, or backwards compatible otherwise
    :param all_language_codes: the set of existing language codes, or None
    :raise ValidationError: if the schema is invalid.
    """
    if root_schema is None:
        if path:
            raise ValidationError('missing root schema', path)
        root_schema = schema
    if all_language_codes is None:
        all_language_codes = get_language_codes()
    if path is None:
        path = []
    if not isinstance(schema, dict):
        raise ValidationError('invalid schema (must be dict)', path)
    if 'type' not in schema:
        raise ValidationError('invalid schema (must contain type)', path)
    if not isinstance(schema['type'], str):
        raise ValidationError('invalid schema (type must be str)', path)
    if 'title' not in schema:
        raise ValidationError('invalid schema (must contain title)', path)
    _validate_title_and_tooltip_in_schema(schema, path, all_language_codes=all_language_codes, strict=strict)
    if 'style' in schema and not isinstance(schema['style'], str) and not (isinstance(schema['style'], dict) and all(isinstance(key, str) for key in schema['style'].keys()) and all(isinstance(value, str) or value is None for value in schema['style'].values())):
        raise ValidationError('style must be a string or a dict of string keys and string or null values', path)
    if 'conditions' in schema:
        if parent_conditions is not None:
            if not isinstance(schema['conditions'], list):
                raise ValidationError('conditions must be a list', path)
            for i, condition in enumerate(schema['conditions']):
                parent_conditions.append((path + [str(i)], condition))
        else:
            raise ValidationError('only schemas with an object parent may contain conditions', path)
    if not isinstance(schema.get('may_copy', True), bool):
        raise ValidationError('may_copy must be bool', path)

    if path == [] and schema['type'] != 'object':
        raise ValidationError('invalid schema (root must be an object)', path)
    if schema['type'] == 'array':
        return _validate_array_schema(schema, path, invalid_template_action_ids=invalid_template_action_ids, strict=strict, all_language_codes=all_language_codes, root_schema=root_schema)
    elif schema['type'] == 'object':
        return _validate_object_schema(schema, path, invalid_template_action_ids=invalid_template_action_ids, strict=strict, all_language_codes=all_language_codes, root_schema=root_schema)
    elif schema['type'] == 'text':
        return _validate_text_schema(schema, path, all_language_codes=all_language_codes, strict=strict)
    elif schema['type'] == 'datetime':
        return _validate_datetime_schema(schema, path, all_language_codes=all_language_codes, strict=strict)
    elif schema['type'] == 'bool':
        return _validate_bool_schema(schema, path, all_language_codes=all_language_codes, strict=strict)
    elif schema['type'] == 'quantity':
        return _validate_quantity_schema(schema, path, all_language_codes=all_language_codes, strict=strict, root_schema=root_schema)
    elif schema['type'] == 'sample':
        return _validate_sample_schema(schema, path, all_language_codes=all_language_codes, strict=strict)
    elif schema['type'] == 'measurement':
        return _validate_measurement_schema(schema, path, all_language_codes=all_language_codes, strict=strict)
    elif schema['type'] == 'object_reference':
        return _validate_object_reference_schema(schema, path, all_language_codes=all_language_codes, strict=strict)
    elif schema['type'] == 'tags':
        return _validate_tags_schema(schema, path, strict=strict)
    elif schema['type'] == 'hazards':
        return _validate_hazards_schema(schema, path, all_language_codes=all_language_codes, strict=strict)
    elif schema['type'] == 'user':
        return _validate_user_schema(schema, path, all_language_codes=all_language_codes, strict=strict)
    elif schema['type'] == 'plotly_chart':
        return _validate_plotly_chart_schema(schema, path, all_language_codes=all_language_codes, strict=strict)
    elif schema['type'] == 'timeseries':
        return _validate_timeseries_schema(schema, path, all_language_codes=all_language_codes, strict=strict)
    elif schema['type'] == 'file':
        return _validate_file_schema(schema, path, all_language_codes=all_language_codes, strict=strict)
    else:
        raise ValidationError('invalid type', path)


def _validate_title_and_tooltip_in_schema(
    schema: typing.Dict[str, typing.Any],
    path: typing.List[str],
    *,
    all_language_codes: typing.Set[str],
    strict: bool
) -> None:
    for key in ['title', 'tooltip']:
        if key == 'tooltip' and key not in schema:
            continue
        if not isinstance(schema[key], str) and not isinstance(schema[key], dict):
            raise ValidationError(f'{key} must be str or dict', path)
        if isinstance(schema[key], dict):
            if 'en' not in schema[key]:
                raise ValidationError(f'{key} must include an english translation', path)
            for lang_code in schema[key].keys():
                if lang_code not in all_language_codes:
                    raise ValidationError(f'{key} must only contain known languages', path)
            for title_text in schema[key].values():
                if not isinstance(title_text, str):
                    raise ValidationError(f'{key} must only contain text', path)
                if strict or key == 'tooltip':
                    if not title_text:
                        raise ValidationError(f'{key} must not be empty', path)
                    if re.match(r'^\s+$', title_text):
                        raise ValidationError(f'{key} must not be whitespace only', path)
        elif strict or key == 'tooltip':
            title_text = schema[key]
            if not title_text:
                raise ValidationError(f'{key} must not be empty', path)
            if re.match(r'^\s+$', title_text):
                raise ValidationError(f'{key} must not be whitespace only', path)


def _validate_note_in_schema(
        schema: typing.Dict[str, typing.Any],
        path: typing.List[str],
        *,
        all_language_codes: typing.Set[str],
        strict: bool
) -> None:
    if 'note' in schema and not isinstance(schema['note'], str) and not isinstance(schema['note'], dict):
        raise ValidationError('note must be str or dict', path)
    if 'note' in schema and isinstance(schema['note'], str):
        note_text = schema['note']
        if strict and not note_text:
            raise ValidationError('note must not be empty', path)
        if strict and re.match(r'^\s+$', note_text):
            raise ValidationError('note must not be whitespace only', path)
    if 'note' in schema and isinstance(schema['note'], dict):
        if 'en' not in schema['note']:
            raise ValidationError('note must include an english translation', path)
        for lang_code in schema['note'].keys():
            if lang_code not in all_language_codes:
                raise ValidationError('note must only contain known languages', path)
        for note_text in schema['note'].values():
            if not isinstance(note_text, str):
                raise ValidationError('note must only contain text', path)
            if strict and not note_text:
                raise ValidationError('note text must not be empty', path)
            if strict and re.match(r'^\s+$', note_text):
                raise ValidationError('note text must not be whitespace only', path)


def _validate_hazards_schema(
        schema: typing.Dict[str, typing.Any],
        path: typing.List[str],
        *,
        all_language_codes: typing.Set[str],
        strict: bool
) -> None:
    """
    Validate the given GHS hazards schema and raise a ValidationError if it is invalid.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :param all_language_codes: the set of existing language codes
    :raise ValidationError: if the schema is invalid.
    """
    valid_keys = {'type', 'title', 'note', 'dataverse_export', 'scicat_export', 'may_copy', 'style', 'tooltip'}
    required_keys = {'type', 'title'}
    schema_keys = set(schema.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError(f'unexpected keys in schema: {invalid_keys}', path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError(f'missing keys in schema: {missing_keys}', path)
    if 'dataverse_export' in schema and not isinstance(schema['dataverse_export'], bool):
        raise ValidationError('dataverse_export must be True or False', path)
    if 'scicat_export' in schema and not isinstance(schema['scicat_export'], bool):
        raise ValidationError('scicat_export must be True or False', path)
    if path != ['hazards']:
        raise ValidationError('GHS hazards must be a top-level entry named "hazards"', path)

    _validate_note_in_schema(schema, path, all_language_codes=all_language_codes, strict=strict)


def _validate_array_schema(
        schema: typing.Dict[str, typing.Any],
        path: typing.List[str],
        *,
        root_schema: typing.Dict[str, typing.Any],
        invalid_template_action_ids: typing.Sequence[int] = (),
        strict: bool,
        all_language_codes: typing.Set[str]
) -> None:
    """
    Validates the given array schema and raises a ValidationError if it is invalid.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :param invalid_template_action_ids: IDs of actions that may not be used as templates to prevent recursion
    :param strict: whether the schema should be evaluated in strict mode, or backwards compatible otherwise
    :param all_language_codes: the set of existing language codes
    :raise ValidationError: if the schema is invalid.
    """
    valid_keys = {'type', 'title', 'items', 'style', 'tooltip', 'minItems', 'maxItems', 'defaultItems', 'default', 'may_copy', 'conditions', 'note'}
    required_keys = {'type', 'title', 'items'}
    schema_keys = set(schema.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError(f'unexpected keys in schema: {invalid_keys}', path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError(f'missing keys in schema: {missing_keys}', path)
    has_min_items = False
    if 'minItems' in schema:
        if not isinstance(schema['minItems'], int):
            raise ValidationError('minItems must be an integer', path)
        elif schema['minItems'] < 0:
            raise ValidationError('minItems must not be negative', path)
        else:
            has_min_items = True
    has_max_items = False
    if 'maxItems' in schema:
        if not isinstance(schema['maxItems'], int):
            raise ValidationError('maxItems must be an integer', path)
        elif schema['maxItems'] < 0:
            raise ValidationError('maxItems must not be negative', path)
        else:
            has_max_items = True
    has_default_items = False
    if 'defaultItems' in schema:
        if not isinstance(schema['defaultItems'], int):
            raise ValidationError('defaultItems must be an integer', path)
        elif schema['defaultItems'] < 0:
            raise ValidationError('defaultItems must not be negative', path)
        else:
            has_default_items = True
    if has_min_items and has_max_items:
        if schema['minItems'] > schema['maxItems']:
            raise ValidationError('minItems must be less than or equal to maxItems', path)
    if has_min_items and has_default_items:
        if schema['minItems'] > schema['defaultItems']:
            raise ValidationError('minItems must be less than or equal to defaultItems', path)
    if has_default_items and has_max_items:
        if schema['defaultItems'] > schema['maxItems']:
            raise ValidationError('defaultItems must be less than or equal to maxItems', path)
    validate_schema(schema['items'], path + ['[?]'], invalid_template_action_ids=invalid_template_action_ids, strict=strict, all_language_codes=all_language_codes, root_schema=root_schema)
    if 'default' in schema:
        if has_default_items:
            raise ValidationError('default and defaultItems are mutually exclusive', path)
        validate(schema['default'], schema, path + ['(default)'], strict=strict)
    _validate_note_in_schema(schema, path, all_language_codes=all_language_codes, strict=strict)


def _validate_tags_schema(
        schema: typing.Dict[str, typing.Any],
        path: typing.List[str],
        *,
        strict: bool
) -> None:
    """
    Validates the given tags schema and raises a ValidationError if it is invalid.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :param strict: whether the schema should be evaluated in strict mode, or backwards compatible otherwise
    :raise ValidationError: if the schema is invalid.
    """
    valid_keys = {'type', 'title', 'default', 'dataverse_export', 'scicat_export', 'may_copy', 'style', 'tooltip'}
    required_keys = {'type', 'title'}
    schema_keys = set(schema.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError(f'unexpected keys in schema: {invalid_keys}', path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError(f'missing keys in schema: {missing_keys}', path)
    if 'default' in schema:
        validate({'_type': 'tags', 'tags': schema['default']}, schema, path + ['(default)'], strict=strict)
    if 'dataverse_export' in schema and not isinstance(schema['dataverse_export'], bool):
        raise ValidationError('dataverse_export must be True or False', path)
    if 'scicat_export' in schema and not isinstance(schema['scicat_export'], bool):
        raise ValidationError('scicat_export must be True or False', path)
    if path != ['tags']:
        raise ValidationError('Tags must be a top-level entry named "tags"', path)


def _validate_property_name(property_name: str, strict: bool, path: typing.List[str]) -> None:
    property_name_valid = True
    if not property_name:
        property_name_valid = False
    if '__' in property_name:
        property_name_valid = False
    if strict and property_name_valid:
        # property name may only consist of ascii characters, digits and underscores
        if not all(c in (string.ascii_letters + string.digits + '_') for c in property_name):
            property_name_valid = False
        # property name must start with a character
        if property_name[0] not in string.ascii_letters:
            property_name_valid = False
        # property name must not end with an underscore
        if property_name.endswith('_'):
            property_name_valid = False
    if not property_name_valid:
        raise ValidationError(f'invalid property name: {property_name}', path)


def _validate_object_schema(
        schema: typing.Dict[str, typing.Any],
        path: typing.List[str],
        *,
        root_schema: typing.Dict[str, typing.Any],
        invalid_template_action_ids: typing.Sequence[int] = (),
        strict: bool,
        all_language_codes: typing.Set[str]
) -> None:
    """
    Validates the given object schema and raises a ValidationError if it is invalid.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :param invalid_template_action_ids: IDs of actions that may not be used as templates to prevent recursion
    :param strict: whether the schema should be evaluated in strict mode, or backwards compatible otherwise
    :param all_language_codes: the set of existing language codes
    :raise ValidationError: if the schema is invalid.
    """
    try:
        substitute_templates(schema, invalid_template_action_ids)
    except ActionDoesNotExistError:
        raise ValidationError('schema template action does not exist', path)
    except InvalidNumberError:
        raise ValidationError('template must be an integer', path)
    except RecursiveTemplateError:
        raise ValidationError('template must not recursively include itself', path)
    except InvalidTemplateIDError:
        raise ValidationError('template must be the ID of a template action', path)

    if schema.get('template') is not None:
        invalid_template_action_ids = list(invalid_template_action_ids) + [typing.cast(int, schema['template'])]

    valid_keys = {'type', 'title', 'properties', 'propertyOrder', 'required', 'default', 'may_copy', 'style', 'template', 'recipes', 'note', 'show_more', 'workflow_show_more'}
    if not path:
        # the top level object may contain a list of properties to be displayed in a table of objects
        valid_keys.add('displayProperties')
        valid_keys.add('batch')
        valid_keys.add('batch_name_format')
        valid_keys.add('notebookTemplates')
        valid_keys.add('workflow_view')
        valid_keys.add('workflow_views')
    if path:
        # the top level object must not have any conditions
        valid_keys.add('conditions')
    required_keys = {'type', 'title', 'properties'}
    schema_keys = set(schema.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError(f'unexpected keys in schema: {invalid_keys}', path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError(f'missing keys in schema: {missing_keys}', path)

    if not isinstance(schema['properties'], dict):
        raise ValidationError('properties must be dict', path)
    property_conditions: typing.List[typing.Tuple[typing.List[str], typing.Dict[str, typing.Any]]] = []
    property_calculations: typing.List[typing.Tuple[typing.List[str], typing.Dict[str, typing.Any]]] = []
    property_schemas = {}
    for property_name, property_schema in schema['properties'].items():
        _validate_property_name(property_name, strict, path)

        validate_schema(
            property_schema,
            path + [property_name],
            parent_conditions=property_conditions,
            invalid_template_action_ids=invalid_template_action_ids,
            strict=strict,
            all_language_codes=all_language_codes,
            root_schema=root_schema
        )
        property_schemas[property_name] = property_schema
    for condition_path, condition in property_conditions:
        if not isinstance(condition, dict) or not isinstance(condition.get('type'), str):
            raise ValidationError('condition must be a dict containing the key type', condition_path)
        validate_condition_schema(condition, property_schemas, condition_path)
    for calculation_path, calculation in property_calculations:
        calculations.validate_calculation(calculation, root_schema, calculation_path)

    if 'required' in schema:
        if not isinstance(schema['required'], list):
            raise ValidationError('required must be list', path)
        for i, property_name in enumerate(schema['required']):
            if property_name not in schema['properties']:
                raise ValidationError(f'unknown required property: {property_name}', path)
            if property_name in schema['required'][:i]:
                raise ValidationError(f'duplicate required property: {property_name}', path)
            # the name property in the root object is always required and may not be conditional
            if schema['properties'][property_name].get('conditions') and property_name == 'name' and path == []:
                raise ValidationError(f'conditional required property: {property_name}', path)

    if 'hazards' in schema['properties'] and schema['properties']['hazards']['type'] == 'hazards' and 'hazards' not in schema.get('required', []):
        raise ValidationError('GHS hazards may not be optional', path)

    if not path:
        if 'name' not in schema['properties'] or schema['properties']['name']['type'] != 'text':
            raise ValidationError('Schema must include a text property "name"', path)
        if schema['properties']['name'].get('multiline') or schema['properties']['name'].get('markdown'):
            raise ValidationError('Object name must not be multiline or markdown text', path)
        if 'required' not in schema or 'name' not in schema['required']:
            raise ValidationError('"name" must be a required property for the root object', path)

    if 'propertyOrder' in schema:
        if not isinstance(schema['propertyOrder'], list):
            raise ValidationError('propertyOrder must be list', path)
        for i, property_name in enumerate(schema['propertyOrder']):
            if property_name not in schema['properties']:
                raise ValidationError(f'unknown propertyOrder property: {property_name}', path)
            if property_name in schema['propertyOrder'][:i]:
                raise ValidationError(f'duplicate propertyOrder property: {property_name}', path)

    if 'default' in schema:
        validate(schema['default'], schema, path + ['(default)'], strict=strict)

    if 'displayProperties' in schema:
        if not isinstance(schema['displayProperties'], list):
            raise ValidationError('displayProperties must be list', path)
        for i, property_name in enumerate(schema['displayProperties']):
            if property_name not in schema['properties']:
                raise ValidationError(f'unknown display property: {property_name}', path)
            if property_name in schema['displayProperties'][:i]:
                raise ValidationError(f'duplicate displayProperties property: {property_name}', path)

    if 'batch' in schema:
        if not isinstance(schema['batch'], bool):
            raise ValidationError('batch must be bool', path)

    if 'batch_name_format' in schema:
        if not schema.get('batch', False):
            raise ValidationError('batch must be True for batch_name_format to be set', path)
        if not isinstance(schema['batch_name_format'], str):
            raise ValidationError('batch_name_format must be a string', path)
        try:
            schema['batch_name_format'].format(1)
        except (ValueError, KeyError):
            raise ValidationError('invalid batch_name_format', path)

    if 'notebookTemplates' in schema:
        _validate_notebook_templates(schema['notebookTemplates'])

    if 'recipes' in schema:
        if not isinstance(schema['recipes'], list):
            raise ValidationError('recipes must be list', path)
        for i, recipe in enumerate(schema['recipes']):
            if 'name' not in recipe:
                raise ValidationError('missing recipe name', path + ['(recipes)', f'[{i}]'])
            if not isinstance(recipe['name'], str) and not isinstance(recipe['name'], dict):
                raise ValidationError('recipe name must be str or dict', path + ['(recipes)', f'[{i}]'])
            if isinstance(recipe['name'], dict):
                if 'en' not in recipe['name']:
                    raise ValidationError('recipe name must include an english translation', path + ['(recipes)', f'[{i}]'])
                for lang_code in recipe['name'].keys():
                    if lang_code not in all_language_codes:
                        raise ValidationError('recipe name must only contain known languages', path + ['(recipes)', f'[{i}]'])
                for name_text in recipe['name'].values():
                    if not isinstance(name_text, str):
                        raise ValidationError('recipe name must only contain text', path + ['(recipes)', f'[{i}]'])
            if 'property_values' not in recipe:
                raise ValidationError('missing property_values', path + ['(recipes)', f'[{i}]'])
            for property_name in recipe['property_values']:
                if property_name not in schema['properties'].keys():
                    raise ValidationError(f'unknown property: {property_name}', path + ['(recipes)', f'[{i}]'])
                if schema['properties'][property_name]['type'] not in ['text', 'quantity', 'datetime', 'bool']:
                    raise ValidationError('unsupported type in recipe', path + ['(recipes)', f'[{i}]', property_name])
                if recipe['property_values'][property_name] is not None:
                    validate(recipe['property_values'][property_name], schema['properties'][property_name], path + ['(recipes)', f'[{i}]', property_name], strict=strict)
                    if recipe['property_values'][property_name]['_type'] == 'quantity' and \
                       ((isinstance(schema['properties'][property_name]['units'], str) and recipe['property_values'][property_name]['units'] != schema['properties'][property_name]['units']) or
                       (recipe['property_values'][property_name]['units'] not in schema['properties'][property_name]['units'])):
                        raise ValidationError(f'Invalid unit {recipe["property_values"][property_name]["units"]}, allowed unit(s): {schema["properties"][property_name]["units"]}', path + ['(recipes)', f'[{i}]', property_name])
                elif schema['properties'][property_name]['type'] == 'bool':
                    raise ValidationError('recipe values for type \'bool\' must not be None', path + ['(recipes)', f'[{i}]', property_name])

    for show_more_key in ['show_more', 'workflow_show_more']:
        if show_more_key in schema:
            if not isinstance(schema[show_more_key], list):
                raise ValidationError(f'{show_more_key} must be list', path)
            for property_name in schema[show_more_key]:
                if property_name not in schema['properties'].keys():
                    raise ValidationError(f'unknown property: {property_name}', path)

    if 'workflow_view' in schema and 'workflow_views' in schema:
        raise ValidationError('workflow_view and workflow_views are mutually exclusive', path)
    if 'workflow_view' in schema:
        workflow_views = [schema['workflow_view']]
    else:
        workflow_views = schema.get('workflow_views', [])
    for workflow_index, workflow_view in enumerate(workflow_views):
        id_keys = ['referencing_action_id', 'referenced_action_id', 'referencing_action_type_id', 'referenced_action_type_id']
        workflow_valid_keys = set(id_keys + ['title', 'show_action_info', 'sorting_properties', 'recursion_filters', 'referenced_filter_operator', 'referencing_filter_operator'])
        if not isinstance(workflow_view, dict):
            raise ValidationError(f'workflow_view {workflow_index} must be a dict', path)
        for key in workflow_view.keys():
            if key not in workflow_valid_keys:
                raise ValidationError(f'invalid key in workflow_view {workflow_index}: {key}', path)
        for key in id_keys:
            if key in workflow_view and not (
                workflow_view[key] is None or
                type(workflow_view[key]) is int or
                type(workflow_view[key]) is list and all(
                    type(action_type_id) is int for action_type_id in workflow_view[key]
                )
            ):
                raise ValidationError(f'{key} in workflow_view {workflow_index} must be int, None or a list of ints', path)
        if 'title' in workflow_view:
            _validate_title_and_tooltip_in_schema(workflow_view, path, all_language_codes=all_language_codes, strict=strict)
        if 'show_action_info' in workflow_view and not isinstance(workflow_view['show_action_info'], bool):
            raise ValidationError(f'show_action_info in workflow_view {workflow_index} must be bool', path)
        if 'sorting_properties' in workflow_view and not isinstance(workflow_view['sorting_properties'], list):
            raise ValidationError(f'sorting_properties in workflow_view {workflow_index} must be list', path)
        for property_name in workflow_view.get('sorting_properties', []):
            _validate_property_name(property_name, False, path)
        _validate_filter_operators(
            mapping=workflow_view,
            filter_operator_keys=('referencing_filter_operator', 'referenced_filter_operator'),
            path=path
        )
        if 'recursion_filters' in workflow_view:
            recursion_id_keys = ['referenced_action_id', 'referenced_action_type_id', 'referencing_action_id', 'referencing_action_type_id']
            recursion_valid_keys = set(id_keys + ['max_depth', 'referenced_filter_operator', 'referencing_filter_operator'])
            recursion_filters = workflow_view['recursion_filters']
            if not isinstance(recursion_filters, dict):
                raise ValidationError(f'recursion_filters in workflow_view {workflow_index} must be a dict', path)
            for key in recursion_filters.keys():
                if key not in recursion_valid_keys:
                    raise ValidationError(f'invalid key in recursion_filters in workflow_view {workflow_index}: {key}', path)
            for key in recursion_id_keys:
                if key in recursion_filters and not (
                    recursion_filters[key] is None or
                    type(recursion_filters[key]) is int or
                    type(recursion_filters[key]) is list and all(
                        type(action_type_id) is int for action_type_id in recursion_filters[key]
                    )
                ):
                    raise ValidationError(f'{key} in recursion_filters in workflow_view {workflow_index} must be int, None or a list of ints', path)
            if 'max_depth' in recursion_filters and not isinstance(recursion_filters['max_depth'], int):
                raise ValidationError(f'max_depth in recursion_filters in workflow_view {workflow_index} must be int', path)
            if 'max_depth' in recursion_filters and recursion_filters['max_depth'] <= 0:
                raise ValidationError(f'max_depth in recursion_filters in workflow_view {workflow_index} must not be negative', path)
            _validate_filter_operators(
                mapping=recursion_filters,
                filter_operator_keys=('referencing_filter_operator', 'referenced_filter_operator'),
                path=path
            )

    _validate_note_in_schema(schema, path, all_language_codes=all_language_codes, strict=strict)


def _validate_filter_operators(
        *,
        mapping: typing.Dict[str, typing.Any],
        filter_operator_keys: typing.Sequence[str],
        valid_filter_operators: typing.Sequence[str] = ('and', 'or'),
        default_filter_operator: str = 'and',
        path: typing.List[str]
) -> None:
    for filter_operator_key in filter_operator_keys:
        if mapping.get(filter_operator_key, default_filter_operator) not in valid_filter_operators:
            raise ValidationError(f'{filter_operator_key} must be one of:' + ', '.join(map(repr, valid_filter_operators)), path)


def _validate_text_schema(
        schema: typing.Dict[str, typing.Any],
        path: typing.List[str],
        *,
        all_language_codes: typing.Set[str],
        strict: bool
) -> None:
    """
    Validates the given text object schema and raises a ValidationError if it is invalid.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :param all_language_codes: the set of existing language codes
    :raise ValidationError: if the schema is invalid.
    """
    valid_keys = {'type', 'title', 'default', 'minLength', 'maxLength', 'choices', 'pattern', 'multiline', 'markdown', 'note', 'placeholder', 'dataverse_export', 'scicat_export', 'languages', 'conditions', 'may_copy', 'style', 'tooltip'}
    schema_keys = set(schema.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError(f'unexpected keys in schema: {invalid_keys}', path)

    if 'languages' in schema:
        if schema['languages'] != 'all':
            if not isinstance(schema['languages'], list) or len(schema['languages']) == 0:
                raise ValidationError('languages must be a non-empty list of known language codes or "all"', path)
            for language in schema['languages']:
                if not isinstance(language, str) or language not in all_language_codes:
                    raise ValidationError('languages must be a list of known language codes or "all"', path)
            allowed_language_codes = set(schema['languages'])
        else:
            allowed_language_codes = all_language_codes
    else:
        allowed_language_codes = {'en'}
    if 'placeholder' in schema and 'choices' in schema:
        raise ValidationError('placeholder cannot be used together with choices', path)
    if 'placeholder' in schema and not isinstance(schema['placeholder'], str) and not isinstance(schema['placeholder'], dict):
        raise ValidationError('placeholder must be str or dict', path)
    if 'placeholder' in schema and isinstance(schema['placeholder'], dict):
        if 'en' not in schema['placeholder']:
            raise ValidationError('placeholder must include an english translation', path)
        for lang_code in schema['placeholder'].keys():
            if lang_code not in all_language_codes:
                raise ValidationError('placeholder must only contain known languages', path)
        for placeholder_text in schema['placeholder'].values():
            if not isinstance(placeholder_text, str):
                raise ValidationError('placeholder must only contain text', path)
    if 'minLength' in schema and (not isinstance(schema['minLength'], int) or isinstance(schema['minLength'], bool)):
        raise ValidationError('minLength must be int', path)
    if 'maxLength' in schema and (not isinstance(schema['maxLength'], int) or isinstance(schema['maxLength'], bool)):
        raise ValidationError('maxLength must be int', path)
    if 'minLength' in schema and schema['minLength'] < 0:
        raise ValidationError('minLength must not be negative', path)
    if 'maxLength' in schema and schema['maxLength'] < 0:
        raise ValidationError('maxLength must not be negative', path)
    if 'minLength' in schema and 'maxLength' in schema and schema['maxLength'] < schema['minLength']:
        raise ValidationError('maxLength must not be less than minLength', path)
    if 'choices' in schema and not isinstance(schema['choices'], list):
        raise ValidationError('choices must be list', path)
    if 'choices' in schema and not schema['choices']:
        raise ValidationError('choices must not be empty', path)
    if 'choices' in schema:
        choice_type = None
        for i, choice in enumerate(schema['choices']):
            if not isinstance(choice, str) and not isinstance(choice, dict):
                raise ValidationError('choice must be str or dict', path + [str(i)])
            if choice_type is not None and type(choice) is not choice_type:
                raise ValidationError('choices must be either all str or all dict', path + [str(i)])
            choice_type = type(choice)
            if isinstance(choice, dict):
                if 'en' not in choice:
                    raise ValidationError('choice must include an english translation', path + [str(i)])
                for lang_code in choice.keys():
                    if lang_code not in all_language_codes:
                        raise ValidationError('choice must only contain known languages', path + [str(i)])
                for choice_text in choice.values():
                    if not isinstance(choice_text, str):
                        raise ValidationError('choice must only contain text', path + [str(i)])
                    if choice_text.isspace():
                        raise ValidationError('choice must contain more than whitespace', path + [str(i)])
            else:
                if choice.isspace():
                    raise ValidationError('choice must contain more than whitespace', path + [str(i)])
    if 'default' in schema and not isinstance(schema['default'], str) and not isinstance(schema['default'], dict):
        raise ValidationError('default must be str or dict', path)
    if 'choices' in schema and 'default' in schema:
        if schema['default'] not in schema['choices']:
            raise ValidationError('default must only contain a valid choice', path)
    elif 'default' in schema and isinstance(schema['default'], dict):
        for lang_code in schema['default'].keys():
            if lang_code not in allowed_language_codes:
                raise ValidationError('default must only contain allowed languages', path)
        for default_text in schema['default'].values():
            if not isinstance(default_text, str):
                raise ValidationError('default must only contain text', path)
    if 'pattern' in schema and not isinstance(schema['pattern'], str):
        raise ValidationError('pattern must be str', path)
    if 'pattern' in schema:
        try:
            re.compile(schema['pattern'])
        except re.error:
            raise ValidationError('pattern is no valid regular expression', path)
    if 'multiline' in schema and not isinstance(schema['multiline'], bool):
        raise ValidationError('multiline must be bool', path)
    if 'markdown' in schema and not isinstance(schema['markdown'], bool):
        raise ValidationError('markdown must be bool', path)
    if schema.get('markdown') and schema.get('multiline'):
        raise ValidationError('text may not both be multiline and markdown', path)
    if 'choices' in schema and schema.get('multiline'):
        raise ValidationError('text may not both be multiline and have choices', path)
    if 'choices' in schema and schema.get('markdown'):
        raise ValidationError('text may not both be markdown and have choices', path)
    if 'dataverse_export' in schema and not isinstance(schema['dataverse_export'], bool):
        raise ValidationError('dataverse_export must be True or False', path)
    if 'dataverse_export' in schema and not schema['dataverse_export'] and path == ['name']:
        raise ValidationError('dataverse_export must be True for the object name', path)
    if 'scicat_export' in schema and not isinstance(schema['scicat_export'], bool):
        raise ValidationError('scicat_export must be True or False', path)
    if 'scicat_export' in schema and not schema['scicat_export'] and path == ['name']:
        raise ValidationError('scicat_export must be True for the object name', path)
    _validate_note_in_schema(schema, path, all_language_codes=all_language_codes, strict=strict)


def _validate_datetime_schema(
        schema: typing.Dict[str, typing.Any],
        path: typing.List[str],
        *,
        all_language_codes: typing.Set[str],
        strict: bool
) -> None:
    """
    Validates the given datetime object schema and raises a ValidationError if it is invalid.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :param all_language_codes: the set of existing language codes
    :raise ValidationError: if the schema is invalid.
    """
    valid_keys = {'type', 'title', 'default', 'note', 'dataverse_export', 'scicat_export', 'conditions', 'may_copy', 'style', 'tooltip'}
    schema_keys = set(schema.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError(f'unexpected keys in schema: {invalid_keys}', path)

    if 'default' in schema:
        if isinstance(schema['default'], str):
            try:
                datetime.datetime.strptime(schema['default'], '%Y-%m-%d %H:%M:%S')
            except ValueError:
                raise ValidationError('invalid default value', path)
        elif schema['default'] is not None:
            raise ValidationError('default must be str or null', path)
    if 'dataverse_export' in schema and not isinstance(schema['dataverse_export'], bool):
        raise ValidationError('dataverse_export must be True or False', path)
    if 'scicat_export' in schema and not isinstance(schema['scicat_export'], bool):
        raise ValidationError('scicat_export must be True or False', path)
    _validate_note_in_schema(schema, path, all_language_codes=all_language_codes, strict=strict)


def _validate_bool_schema(
        schema: typing.Dict[str, typing.Any],
        path: typing.List[str],
        *,
        all_language_codes: typing.Set[str],
        strict: bool
) -> None:
    """
    Validates the given boolean object schema and raises a ValidationError if it is invalid.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :raise ValidationError: if the schema is invalid.
    """
    valid_keys = {'type', 'title', 'default', 'note', 'dataverse_export', 'scicat_export', 'conditions', 'may_copy', 'style', 'tooltip'}
    schema_keys = set(schema.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError(f'unexpected keys in schema: {invalid_keys}', path)

    if 'default' in schema and not isinstance(schema['default'], bool):
        raise ValidationError('default must be bool', path)
    if 'dataverse_export' in schema and not isinstance(schema['dataverse_export'], bool):
        raise ValidationError('dataverse_export must be True or False', path)
    if 'scicat_export' in schema and not isinstance(schema['scicat_export'], bool):
        raise ValidationError('scicat_export must be True or False', path)
    _validate_note_in_schema(schema, path, all_language_codes=all_language_codes, strict=strict)


def _validate_quantity_schema(
        schema: typing.Dict[str, typing.Any],
        path: typing.List[str],
        *,
        root_schema: typing.Dict[str, typing.Any],
        all_language_codes: typing.Set[str],
        strict: bool
) -> None:
    """
    Validates the given quantity object schema and raises a ValidationError if it is invalid.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :param all_language_codes: the set of existing language codes
    :param strict: whether the schema should be evaluated in strict mode, or backwards compatible otherwise
    :raise ValidationError: if the schema is invalid.
    """
    valid_keys = {'type', 'title', 'units', 'default', 'note', 'placeholder', 'dataverse_export', 'scicat_export', 'conditions', 'may_copy', 'style', 'tooltip', 'display_digits', 'min_magnitude', 'max_magnitude', 'calculation'}
    required_keys = {'type', 'title', 'units'}
    schema_keys = set(schema.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError(f'unexpected keys in schema: {invalid_keys}', path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError(f'missing keys in schema: {missing_keys}', path)

    if isinstance(schema['units'], str):
        if not units_are_valid(schema['units']):
            raise ValidationError('invalid units', path)
    elif isinstance(schema['units'], list) and len(schema['units']) > 0:
        dimensionality = None
        pint_units = []
        for unit in schema['units']:
            if not isinstance(unit, str):
                raise ValidationError('units must be string or list of strings', path)
            try:
                quantity = datatypes.Quantity(1.0, units=unit)
            except pint.UndefinedUnitError:
                raise ValidationError('invalid units', path)
            if dimensionality is None:
                dimensionality = quantity.dimensionality
            elif quantity.dimensionality != dimensionality:
                raise ValidationError('units must be for same dimensionality', path)
            if quantity.pint_units in pint_units:
                raise ValidationError('units must not be duplicate', path)
            pint_units.append(quantity.pint_units)
    else:
        raise ValidationError('units must be string or list of strings', path)

    if 'min_magnitude' in schema and not isinstance(schema['min_magnitude'], float) and not isinstance(schema['min_magnitude'], int):
        raise ValidationError('min_magnitude must be float or int', path)
    if 'max_magnitude' in schema and not isinstance(schema['max_magnitude'], float) and not isinstance(schema['max_magnitude'], int):
        raise ValidationError('max_magnitude must be float or int', path)
    if 'min_magnitude' in schema and not math.isfinite(schema['min_magnitude']):
        raise ValidationError('min_magnitude must be a finite number', path)
    if 'max_magnitude' in schema and not math.isfinite(schema['max_magnitude']):
        raise ValidationError('max_magnitude must be a finite number', path)
    if 'min_magnitude' in schema and 'max_magnitude' in schema and schema['min_magnitude'] > schema['max_magnitude']:
        raise ValidationError('max_magnitude must be greater than or equal to min_magnitude', path)
    if 'default' in schema:
        if isinstance(schema['default'], (float, int)):
            if not math.isfinite(schema['default']):
                raise ValidationError('default must be a finite number', path)
            if 'min_magnitude' in schema and schema['min_magnitude'] > schema['default']:
                raise ValidationError('default must be greater than or equal to min_magnitude', path)
            if 'max_magnitude' in schema and schema['max_magnitude'] < schema['default']:
                raise ValidationError('default must be less than or equal to max_magnitude', path)
        elif isinstance(schema['default'], dict):
            _default_quantity = copy.deepcopy(schema['default'])
            if '_type' not in _default_quantity:
                _default_quantity['_type'] = 'quantity'
            validate(_default_quantity, schema, path + ['(default)'])
        else:
            raise ValidationError('default must be float, int or dict', path)
    if 'dataverse_export' in schema and not isinstance(schema['dataverse_export'], bool):
        raise ValidationError('dataverse_export must be True or False', path)
    if 'scicat_export' in schema and not isinstance(schema['scicat_export'], bool):
        raise ValidationError('scicat_export must be True or False', path)
    if 'placeholder' in schema and not isinstance(schema['placeholder'], str) and not isinstance(schema['placeholder'], dict):
        raise ValidationError('placeholder must be str or dict', path)
    if 'placeholder' in schema and isinstance(schema['placeholder'], dict):
        if 'en' not in schema['placeholder']:
            raise ValidationError('placeholder must include an english translation', path)
        for lang_code in schema['placeholder'].keys():
            if lang_code not in all_language_codes:
                raise ValidationError('placeholder must only contain known languages', path)
        for placeholder_text in schema['placeholder'].values():
            if not isinstance(placeholder_text, str):
                raise ValidationError('placeholder must only contain text', path)
    if 'display_digits' in schema and (type(schema['display_digits']) is not int or schema['display_digits'] < 0):
        raise ValidationError('display_digits must be a non-negative int', path)
    if strict:
        if 'display_digits' in schema and schema['display_digits'] > 15:
            raise ValidationError('display_digits must be at most 15', path)
    if 'calculation' in schema:
        calculations.validate_calculation(schema['calculation'], root_schema, path)
    _validate_note_in_schema(schema, path, all_language_codes=all_language_codes, strict=strict)


def _validate_sample_schema(
        schema: typing.Dict[str, typing.Any],
        path: typing.List[str],
        *,
        all_language_codes: typing.Set[str],
        strict: bool
) -> None:
    """
    Validates the given sample object schema and raises a ValidationError if it is invalid.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :param all_language_codes: the set of existing language codes
    :raise ValidationError: if the schema is invalid.
    """
    valid_keys = {'type', 'title', 'note', 'dataverse_export', 'scicat_export', 'conditions', 'may_copy', 'style', 'tooltip'}
    required_keys = {'type', 'title'}
    schema_keys = set(schema.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError(f'unexpected keys in schema: {invalid_keys}', path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError(f'missing keys in schema: {missing_keys}', path)

    if 'dataverse_export' in schema and not isinstance(schema['dataverse_export'], bool):
        raise ValidationError('dataverse_export must be True or False', path)
    if 'scicat_export' in schema and not isinstance(schema['scicat_export'], bool):
        raise ValidationError('scicat_export must be True or False', path)
    _validate_note_in_schema(schema, path, all_language_codes=all_language_codes, strict=strict)


def _validate_measurement_schema(
        schema: typing.Dict[str, typing.Any],
        path: typing.List[str],
        *,
        all_language_codes: typing.Set[str],
        strict: bool
) -> None:
    """
    Validates the given measurement object schema and raises a ValidationError if it is invalid.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :param all_language_codes: the set of existing language codes
    :raise ValidationError: if the schema is invalid.
    """
    valid_keys = {'type', 'title', 'note', 'dataverse_export', 'scicat_export', 'conditions', 'may_copy', 'style', 'tooltip'}
    required_keys = {'type', 'title'}
    schema_keys = set(schema.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError(f'unexpected keys in schema: {invalid_keys}', path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError(f'missing keys in schema: {missing_keys}', path)

    if 'dataverse_export' in schema and not isinstance(schema['dataverse_export'], bool):
        raise ValidationError('dataverse_export must be True or False', path)
    if 'scicat_export' in schema and not isinstance(schema['scicat_export'], bool):
        raise ValidationError('scicat_export must be True or False', path)
    _validate_note_in_schema(schema, path, all_language_codes=all_language_codes, strict=strict)


def _validate_object_reference_schema(
        schema: typing.Dict[str, typing.Any],
        path: typing.List[str],
        *,
        all_language_codes: typing.Set[str],
        strict: bool
) -> None:
    """
    Validates the given object reference object schema and raises a ValidationError if it is invalid.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :param all_language_codes: the set of existing language codes
    :raise ValidationError: if the schema is invalid.
    """
    valid_keys = {'type', 'title', 'note', 'action_type_id', 'action_id', 'filter_operator', 'dataverse_export', 'scicat_export', 'conditions', 'may_copy', 'style', 'tooltip'}
    required_keys = {'type', 'title'}
    schema_keys = set(schema.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError(f'unexpected keys in schema: {invalid_keys}', path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError(f'missing keys in schema: {missing_keys}', path)

    if 'action_type_id' in schema and not (
            schema['action_type_id'] is None or
            type(schema['action_type_id']) is int or
            type(schema['action_type_id']) is list and all(
                type(action_type_id) is int for action_type_id in schema['action_type_id']
            )
    ):
        raise ValidationError('action_type_id must be int, None or a list of ints', path)
    if 'action_id' in schema and not (
            schema['action_id'] is None or
            type(schema['action_id']) is int or
            type(schema['action_id']) is list and all(
                type(action_type_id) is int for action_type_id in schema['action_id']
            )
    ):
        raise ValidationError('action_id must be int, None or a list of ints', path)
    _validate_filter_operators(
        mapping=schema,
        filter_operator_keys=('filter_operator',),
        path=path
    )
    if 'dataverse_export' in schema and not isinstance(schema['dataverse_export'], bool):
        raise ValidationError('dataverse_export must be True or False', path)
    if 'scicat_export' in schema and not isinstance(schema['scicat_export'], bool):
        raise ValidationError('scicat_export must be True or False', path)
    _validate_note_in_schema(schema, path, all_language_codes=all_language_codes, strict=strict)


def _validate_notebook_templates(notebook_templates: typing.Any) -> None:
    """
    Validate the given notebook templates and raise a ValidationError if they are invalid.

    :param notebook_templates: the sampledb object schema
    :raise ValidationError: if the notebook templates are invalid
    """
    if not isinstance(notebook_templates, list):
        raise ValidationError('notebookTemplates must be a list', ['notebookTemplates'])

    for notebook_index, notebook_template in enumerate(notebook_templates):
        path = ['notebookTemplates', str(notebook_index)]
        if not isinstance(notebook_template, dict):
            raise ValidationError('notebook template must be a dict', path)
        valid_keys = {'title', 'url', 'params'}
        required_keys = valid_keys
        schema_keys = set(notebook_template.keys())
        invalid_keys = schema_keys - valid_keys
        if invalid_keys:
            raise ValidationError(f'unexpected keys in notebook template: {invalid_keys}', path)
        missing_keys = required_keys - schema_keys
        if missing_keys:
            raise ValidationError(f'missing keys in notebook template: {missing_keys}', path)
        if not isinstance(notebook_template['title'], str):
            raise ValidationError('notebook template title must be str', path)
        if not isinstance(notebook_template['url'], str):
            raise ValidationError('notebook template url must be str', path)
        if not notebook_template['url'].endswith('.ipynb'):
            raise ValidationError('notebook template url must end with .ipynb', path)
        base_url = 'https://iffjupyter.fz-juelich.de/templates/t/'
        test_url = urllib.parse.urljoin(base_url, notebook_template['url'])
        if not test_url.startswith(base_url) or urllib.parse.urlparse(base_url).netloc != urllib.parse.urlparse(test_url).netloc:
            raise ValidationError('notebook template url must be relative', path)
        if not isinstance(notebook_template['params'], dict):
            raise ValidationError('notebook template params must be a dict', path)
        for param_name, param_value in notebook_template['params'].items():
            if not isinstance(param_name, str):
                raise ValidationError('notebook template param names must be str', path)
            if not isinstance(param_value, str) and not isinstance(param_value, list):
                raise ValidationError('notebook template param values must be str or a list', path)
            if isinstance(param_value, list):
                for param_step in param_value:
                    if not isinstance(param_step, str) and not isinstance(param_step, int):
                        raise ValidationError('notebook template param value steps must be str or int', path)
            if isinstance(param_value, str):
                valid_param_values = {'object_id'}
                if param_value not in valid_param_values:
                    raise ValidationError(f'notebook template param value must be a list or one of {valid_param_values}', path)


def _validate_user_schema(
        schema: typing.Dict[str, typing.Any],
        path: typing.List[str],
        *,
        all_language_codes: typing.Set[str],
        strict: bool
) -> None:
    """
    Validates the given user object schema and raises a ValidationError if it is invalid.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :param all_language_codes: the set of existing language codes
    :raise ValidationError: if the schema is invalid.
    """
    valid_keys = {'type', 'title', 'note', 'dataverse_export', 'scicat_export', 'default', 'conditions', 'may_copy', 'style', 'tooltip'}
    required_keys = {'type', 'title'}
    schema_keys = set(schema.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError(f'unexpected keys in schema: {invalid_keys}', path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError(f'missing keys in schema: {missing_keys}', path)
    if 'dataverse_export' in schema and not isinstance(schema['dataverse_export'], bool):
        raise ValidationError('dataverse_export must be True or False', path)
    if 'scicat_export' in schema and not isinstance(schema['scicat_export'], bool):
        raise ValidationError('scicat_export must be True or False', path)
    if 'default' in schema and (schema['default'] != 'self' and type(schema['default']) is not int):
        raise ValidationError('default must be "self" or int', path)
    _validate_note_in_schema(schema, path, all_language_codes=all_language_codes, strict=strict)


def _validate_plotly_chart_schema(
        schema: typing.Dict[str, typing.Any],
        path: typing.List[str],
        *,
        all_language_codes: typing.Set[str],
        strict: bool
) -> None:
    """
    Validates the given plotly_chart object schema and raises a ValidationError if it is invalid.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :param all_language_codes: the set of existing language codes
    :raise ValidationError: if the schema is invalid.
    """
    valid_keys = {'type', 'title', 'note', 'dataverse_export', 'conditions', 'may_copy', 'style', 'tooltip'}
    schema_keys = schema.keys()
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError(f'unexpected keys in schema: {invalid_keys}', path)
    _validate_note_in_schema(schema, path, all_language_codes=all_language_codes, strict=strict)


def _validate_timeseries_schema(
        schema: typing.Dict[str, typing.Any],
        path: typing.List[str],
        *,
        all_language_codes: typing.Set[str],
        strict: bool
) -> None:
    """
    Validates the given timeseries schema and raises a ValidationError if it is invalid.

    :param schema: the sampledb object schema
    :param path: the path to this subschema
    :param all_language_codes: the set of existing language codes
    :param strict: whether the schema should be evaluated in strict mode, or backwards compatible otherwise
    :raise ValidationError: if the schema is invalid.
    """
    valid_keys = {'type', 'title', 'units', 'note', 'dataverse_export', 'scicat_export', 'conditions', 'may_copy', 'style', 'tooltip', 'display_digits', 'statistics'}
    required_keys = {'type', 'title', 'units'}
    schema_keys = set(schema.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError(f'unexpected keys in schema: {invalid_keys}', path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError(f'missing keys in schema: {missing_keys}', path)

    if isinstance(schema['units'], str):
        if not units_are_valid(schema['units']):
            raise ValidationError('invalid units', path)
    elif isinstance(schema['units'], list) and len(schema['units']) > 0:
        dimensionality = None
        pint_units = []
        for unit in schema['units']:
            if not isinstance(unit, str):
                raise ValidationError('units must be string or list of strings', path)
            try:
                quantity = datatypes.Quantity(1.0, units=unit)
            except pint.UndefinedUnitError:
                raise ValidationError('invalid units', path)
            if dimensionality is None:
                dimensionality = quantity.dimensionality
            elif quantity.dimensionality != dimensionality:
                raise ValidationError('units must be for same dimensionality', path)
            if quantity.pint_units in pint_units:
                raise ValidationError('units must not be duplicate', path)
            pint_units.append(quantity.pint_units)
    else:
        raise ValidationError('units must be string or list of strings', path)

    if 'dataverse_export' in schema and not isinstance(schema['dataverse_export'], bool):
        raise ValidationError('dataverse_export must be True or False', path)
    if 'scicat_export' in schema and not isinstance(schema['scicat_export'], bool):
        raise ValidationError('scicat_export must be True or False', path)
    if 'display_digits' in schema and (type(schema['display_digits']) is not int or schema['display_digits'] < 0):
        raise ValidationError('display_digits must be a non-negative int', path)
    if strict:
        if 'display_digits' in schema and schema['display_digits'] > 15:
            raise ValidationError('display_digits must be at most 15', path)
    if 'statistics' in schema:
        valid_statistics_keys = {'average', 'stddev', 'min', 'max', 'count', 'first', 'last'}
        if not isinstance(schema['statistics'], list):
            raise ValidationError('statistics must be a list of strings', path)
        if not all(statistics_key in valid_statistics_keys for statistics_key in schema['statistics']):
            raise ValidationError(f'statistics keys each have to be one of the following: {valid_statistics_keys}', path)
    _validate_note_in_schema(schema, path, all_language_codes=all_language_codes, strict=strict)


def _validate_file_schema(
        schema: typing.Dict[str, typing.Any],
        path: typing.List[str],
        *,
        all_language_codes: typing.Set[str],
        strict: bool
) -> None:
    """
    Validates the given file object schema and raises a ValidationError if it is invalid.

    :param schema: the sampledb file object schema
    :param path: the path to this subschema
    :param all_language_codes: the set of existing language codes
    :raise ValidationError: if the schema is invalid.
    """
    valid_keys = {'type', 'title', 'note', 'conditions', 'style', 'tooltip', 'extensions', 'preview'}
    required_keys = {'type', 'title'}
    schema_keys = set(schema.keys())
    invalid_keys = schema_keys - valid_keys
    if invalid_keys:
        raise ValidationError(f'unexpected keys in schema: {invalid_keys}', path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError(f'missing keys in schema: {missing_keys}', path)
    if 'extensions' in schema:
        if not isinstance(schema['extensions'], list) or not schema['extensions'] or not all(isinstance(extension, str) and len(extension) > 1 and extension.startswith('.') for extension in schema['extensions']):
            raise ValidationError('extensions must be a list of file extensions', path)
    if 'preview' in schema and not isinstance(schema['preview'], bool):
        raise ValidationError('preview must be a list of file extensions', path)
    _validate_note_in_schema(schema, path, all_language_codes=all_language_codes, strict=strict)



# coding: utf-8
"""

"""

from flask_wtf import FlaskForm
from wtforms import StringField, IntegerField, FieldList, FormField, BooleanField, SelectMultipleField
from wtforms.validators import InputRequired, NumberRange, DataRequired

from ..models import Permissions


class CreateProjectForm(FlaskForm):
    translations = StringField(validators=[DataRequired()])
    categories = SelectMultipleField()


class EditProjectForm(FlaskForm):
    translations = StringField(validators=[DataRequired()])
    categories = SelectMultipleField()


class LeaveProjectForm(FlaskForm):
    pass


class DeleteProjectForm(FlaskForm):
    pass


class RemoveProjectMemberForm(FlaskForm):
    pass


class RemoveProjectGroupForm(FlaskForm):
    pass


class OtherProjectIdForm(FlaskForm):
    project_id = IntegerField(
        validators=[InputRequired()]
    )
    add_user = BooleanField(default=False)


class InviteUserToProjectForm(FlaskForm):
    user_id = IntegerField(validators=[InputRequired()])
    other_project_ids = FieldList(FormField(OtherProjectIdForm), min_entries=0)
    permissions = IntegerField(validators=[InputRequired(), NumberRange(min=Permissions.READ.value, max=Permissions.GRANT.value)])
    add_directly = BooleanField(default=False)


class InviteGroupToProjectForm(FlaskForm):
    group_id = IntegerField(validators=[InputRequired()])


class AddSubprojectForm(FlaskForm):
    child_project_id = IntegerField(validators=[InputRequired()])
    child_can_add_users_to_parent = BooleanField(default=False)


class RemoveSubprojectForm(FlaskForm):
    child_project_id = IntegerField(validators=[InputRequired()])


class ObjectLinkForm(FlaskForm):
    object_id = IntegerField(validators=[InputRequired()])



"""
Module for generation sample/object labels as PDF files.
"""
import collections
import io
import base64
import math
import os
import typing

from PIL import Image
from weasyprint import HTML

import qrcode
import qrcode.image.pil

import flask


DEFAULT_PAPER_FORMAT = 'DIN A4 (Portrait)'
PAGE_SIZES = {
    'DIN A4 (Portrait)': (210, 297),
    'DIN A4 (Landscape)': (297, 210),
    'Letter (Portrait)': (8.5 * 25.4, 11 * 25.4),
    'Letter (Landscape)': (11 * 25.4, 8.5 * 25.4)
}

PAGE_SIZE_KEYS = ['DIN A4 (Portrait)', 'DIN A4 (Landscape)', 'Letter (Portrait)', 'Letter (Landscape)']

HORIZONTAL_LABEL_MARGIN = 10
VERTICAL_LABEL_MARGIN = 10


def _generate_ghs_image_uris() -> typing.List[str]:
    ghs_image_uris = []
    GHS_IMAGE_DIR = os.path.join(os.path.abspath(os.path.dirname(__file__)), '..', 'static', 'sampledb', 'img')
    for i in range(0, 10):
        ghs_image = Image.open(os.path.join(GHS_IMAGE_DIR, f'ghs0{i}.png')).convert('RGBA')
        image_stream = io.BytesIO()
        ghs_image.save(image_stream, format='png')
        image_stream.seek(0)
        ghs_image_uri = 'data:image/png;base64,' + base64.b64encode(image_stream.read()).decode('utf-8')
        ghs_image_uris.append(ghs_image_uri)
    return ghs_image_uris


GHS_IMAGE_URIS = _generate_ghs_image_uris()


def _generate_font_uris() -> typing.Dict[str, str]:
    font_uris = {}
    font_path = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "static", "urw-base35-fonts"))
    for font_file_name in ["NimbusSans-Regular.otf", "NimbusSans-Bold.otf"]:
        with open(os.path.join(font_path, font_file_name), "rb") as font_file:
            font_uri = 'data:font/otf;base64,' + base64.b64encode(font_file.read()).decode('utf-8')
            font_uris[font_file_name] = font_uri
    return font_uris


FONT_URIS = _generate_font_uris()


def create_labels(
        object_id: int,
        object_name: str,
        object_url: str,
        creation_user: str,
        creation_date: str,
        ghs_classes: typing.List[int],
        paper_format: str = DEFAULT_PAPER_FORMAT,
        create_mixed_labels: bool = True,
        create_long_labels: bool = False,
        create_only_qr_codes: bool = False,
        include_qrcode_in_long_labels: bool = False,
        label_width: float = 18,
        label_minimum_height: float = 0,
        label_minimum_width: float = 0,
        qrcode_width: float = 18,
        label_quantity: int = 1,
        ghs_classes_side_by_side: bool = False,
        centered: bool = True,
        only_id_qr_code: bool = False,
        add_label_number: bool = False,
        add_maximum_label_number: bool = False,
        show_id_on_label: bool = True,
        label_dimension: typing.Optional[dict[str, typing.Any]] = None,
        custom_qr_code_texts: typing.Optional[typing.Dict[str, str]] = None
) -> bytes:
    object_specification = {
        object_id: {
            "object_name": object_name,
            "object_url": object_url,
            "creation_user": creation_user,
            "creation_date": creation_date,
            "ghs_classes": ghs_classes
        }
    }

    return create_multiple_labels(
        object_specifications=object_specification,
        include_qrcode_in_long_labels=include_qrcode_in_long_labels,
        paper_format=paper_format,
        quantity=label_quantity,
        label_width=label_width,
        min_label_width=label_minimum_width,
        qr_code_width=qrcode_width,
        min_label_height=label_minimum_height,
        ghs_classes_side_by_side=ghs_classes_side_by_side,
        centered=centered,
        create_mixed_labels=create_mixed_labels,
        create_long_labels=create_long_labels,
        create_only_qr_codes=create_only_qr_codes,
        only_id_qr_code=only_id_qr_code,
        add_label_number=add_label_number,
        add_maximum_label_number=add_maximum_label_number,
        show_id_on_label=show_id_on_label,
        label_dimension=label_dimension,
        fill_single_page=not create_only_qr_codes,
        custom_qr_code_texts=custom_qr_code_texts
    )


def create_multiple_labels(
        object_specifications: typing.Dict[int, typing.Dict[str, typing.Any]],
        quantity: int = 1,
        label_width: float = 18,
        min_label_width: float = 0,
        min_label_height: float = 0,
        qr_code_width: float = 18,
        paper_format: str = DEFAULT_PAPER_FORMAT,
        create_mixed_labels: bool = False,
        create_long_labels: bool = False,
        create_only_qr_codes: bool = False,
        include_qrcode_in_long_labels: bool = False,
        ghs_classes_side_by_side: bool = False,
        centered: bool = True,
        fill_single_page: bool = False,
        only_id_qr_code: bool = False,
        add_label_number: bool = False,
        add_maximum_label_number: bool = False,
        show_id_on_label: bool = True,
        label_dimension: typing.Optional[dict[str, typing.Any]] = None,
        custom_qr_code_texts: typing.Optional[typing.Dict[str, str]] = None
) -> bytes:
    vertical_label_margin = VERTICAL_LABEL_MARGIN / 4
    horizontal_label_margin = HORIZONTAL_LABEL_MARGIN / 4

    paper_width, paper_height = PAGE_SIZES[paper_format]

    tmp_object_specifications_dict = collections.OrderedDict(sorted(object_specifications.items()))
    object_specifications = tmp_object_specifications_dict

    username_list = []
    object_name_list = []
    creation_date_list = []
    object_url_list = []
    hazard_list = []
    sample_code_list = []
    qr_code_uri_list = []
    ghs_amount_list = []
    ghs_width = 9.0

    if create_long_labels:
        text_extra_width_list = []
        tmp_index = 0
        qrcode_width = 12.5
        for object_id in object_specifications:
            username_list.append(object_specifications[object_id]["creation_user"])
            object_name_list.append(object_specifications[object_id]["object_name"])
            creation_date_list.append(object_specifications[object_id]["creation_date"])
            object_url_list.append(object_specifications[object_id]["object_url"])
            hazard_list.append(object_specifications[object_id]["ghs_classes"])
            sample_code_list.append(object_id)

            url = object_url_list[tmp_index]
            image = qrcode.make(url, border=1)
            image_stream = io.BytesIO()
            image.save(image_stream, format='png')
            image_stream.seek(0)
            qr_code_uri_list.append('data:image/png;base64,' + base64.b64encode(image_stream.read()).decode('utf-8'))

            text_extra_width_list.append(len(hazard_list[tmp_index]) * ghs_width)
            if include_qrcode_in_long_labels:
                text_extra_width_list[tmp_index] = text_extra_width_list[tmp_index] + qrcode_width + 1
            tmp_index += 1

        page_amounts: typing.Dict[str, typing.List[int]] = {
            "DIN A4 (Portrait)": [21, 16],
            "DIN A4 (Landscape)": [14, 11],
            "Letter (Portrait)": [19, 15],
            "Letter (Landscape)": [15, 11]
        }

        if include_qrcode_in_long_labels:
            box_height = 12.24
            label_margin = [horizontal_label_margin / 2 + 0.3, horizontal_label_margin + 0.3]
            page_margin = [9.55, 5, 5, 4.9]
            padding_top = 0.45
            padding_right = 0.3
        else:
            box_height = 7.7
            label_margin = [horizontal_label_margin / 2 + 0.544, horizontal_label_margin + 0.3]
            page_margin = [9.9, 5, 5, 5]
            padding_top = 1.0
            padding_right = 0.0

        if fill_single_page:
            if include_qrcode_in_long_labels:
                num_labels = page_amounts[paper_format][1]
            else:
                num_labels = page_amounts[paper_format][0]
        else:
            num_labels = quantity

        object_amount = len(username_list)
        html = flask.render_template(
            "labels/minimal_height_labels.html",
            username_list=username_list,
            object_name_list=object_name_list,
            creation_date_list=creation_date_list,
            hazard_list=hazard_list,
            qrcode_width=qrcode_width,
            ghs_width=ghs_width,
            object_id_list=sample_code_list,
            qr_code_uri_list=qr_code_uri_list,
            object_amount=object_amount,
            include_qrcode=include_qrcode_in_long_labels,
            box_height=box_height,
            paper_width=paper_width,
            paper_height=paper_height,
            num_labels=num_labels,
            GHS_IMAGE_URIS=GHS_IMAGE_URIS,
            min_label_width=min_label_width,
            text_extra_width_list=text_extra_width_list,
            FONT_URIS=FONT_URIS,
            label_margin=label_margin,
            page_margin=page_margin,
            padding_top=padding_top,
            padding_right=padding_right,
        )

    elif create_mixed_labels:
        has_ghs_list = []

        first_box_min_width_list = []
        second_box_min_width_list = []
        third_box_qrcode_box_height_list = []
        third_box_ghs_box_height_list = []
        fourth_box_qrcode_box_height_list = []
        fourth_box_ghs_box_height_list = []
        fifth_box_height_list = []
        fifth_box_qrcode_box_width_list = []
        fifth_box_ghs_box_width_list = []
        fifth_inner_box_height_list = []
        sixth_box_height_list = []
        sixth_inner_box_height_list = []
        outer_box_height_list = []

        group_box_height_list = []
        text_extra_width_list = []
        text_extra_width_list_qr = []

        third_box_width = 17.5
        forth_box_width = 19.55
        fifth_box_width = 39.5
        sixth_box_width = 74.5
        outer_box_width = 200.0
        sixth_box_qrcode_box_width = 19.75
        sixth_box_ghs_box_width = 19.25
        qrcode_width = 12.5

        tmp_index = 0

        for object_id in object_specifications:
            username_list.append(object_specifications[object_id]["creation_user"])
            object_name_list.append(object_specifications[object_id]["object_name"])
            creation_date_list.append(object_specifications[object_id]["creation_date"])
            object_url_list.append(object_specifications[object_id]["object_url"])
            hazard_list.append(object_specifications[object_id]["ghs_classes"])
            sample_code_list.append(object_id)

            url = object_url_list[tmp_index]
            image = qrcode.make(url, border=1)
            image_stream = io.BytesIO()
            image.save(image_stream, format='png')
            image_stream.seek(0)
            qr_code_uri_list.append('data:image/png;base64,' + base64.b64encode(image_stream.read()).decode('utf-8'))

            ghs_height = 17.0 + int((len(hazard_list[tmp_index]) - 1) / 3) * 9

            first_box_min_width_list.append(ghs_width * len(hazard_list[tmp_index]))

            second_box_min_width_list.append(2 + qrcode_width + ghs_width * len(hazard_list[tmp_index]))

            text_extra_width_list.append(len(hazard_list[tmp_index]) * ghs_width)
            text_extra_width_list_qr.append(len(hazard_list[tmp_index]) * ghs_width + qrcode_width)

            fifth_box_height_list.append(max(50.0, 12.0 + math.ceil(ghs_height)))

            if len(hazard_list[tmp_index]) > 0:
                fifth_box_qrcode_box_width_list.append(fifth_box_width / 2.0)
                fifth_box_ghs_box_width_list.append(fifth_box_width / 2.0)
            else:
                fifth_box_qrcode_box_width_list.append(fifth_box_width)
                fifth_box_ghs_box_width_list.append(0)

            sixth_box_height_list.append(max(50.0, 12.0 + math.ceil(ghs_height)))

            ghs_amount_list.append(len(hazard_list[tmp_index]))

            third_box_qrcode_box_height_list.append(18)
            fourth_box_qrcode_box_height_list.append(20.0)

            if ghs_amount_list[tmp_index] > 0:
                has_ghs_list.append(True)

                if ghs_amount_list[tmp_index] <= 2:
                    third_box_ghs_box_height_list.append(12.0 + int((ghs_amount_list[tmp_index] - 1) / 3) * 9)
                    fourth_box_ghs_box_height_list.append(12.0 + int((ghs_amount_list[tmp_index] - 1) / 3) * 9)
                else:
                    third_box_ghs_box_height_list.append(16.75 + int((ghs_amount_list[tmp_index] - 1) / 3) * 9)
                    fourth_box_ghs_box_height_list.append(16.75 + int((ghs_amount_list[tmp_index] - 1) / 3) * 9)

                if ghs_amount_list[tmp_index] == 2 or ghs_amount_list[tmp_index] == 1:
                    third_box_ghs_box_height_list[tmp_index] += 0.3
                    fourth_box_ghs_box_height_list[tmp_index] += 0.3

                fifth_inner_box_height_list.append(max(22.0, 16.25 + int((ghs_amount_list[tmp_index] - 1) / 3) * 9))
                sixth_inner_box_height_list.append(max(22.0, 16.25 + int((ghs_amount_list[tmp_index] - 1) / 3) * 9))

                outer_box_height_list.append(26.0 + 4 * math.floor(len(object_name_list[tmp_index]) * 2.3 / third_box_width) +
                                             int((ghs_amount_list[tmp_index] - 1) / 3) * 9)

                if ghs_amount_list[tmp_index] == 4:
                    third_box_ghs_box_height_list[tmp_index] -= ghs_width / 2.0
                    fourth_box_ghs_box_height_list[tmp_index] -= ghs_width / 2.0
                    fifth_inner_box_height_list[tmp_index] -= ghs_width / 2 - 0.5
                    sixth_inner_box_height_list[tmp_index] -= ghs_width / 2 - 0.5
                elif ghs_amount_list[tmp_index] == 7:
                    third_box_ghs_box_height_list[tmp_index] -= ghs_width / 2.0
                    fourth_box_ghs_box_height_list[tmp_index] -= ghs_width / 2.0
                    fifth_inner_box_height_list[tmp_index] -= ghs_width / 2.0
                    sixth_inner_box_height_list[tmp_index] -= ghs_width / 2.0

                if ghs_amount_list[tmp_index] > 3:
                    fifth_inner_box_height_list[tmp_index] += 0.75
                    sixth_inner_box_height_list[tmp_index] += 0.75
            else:
                has_ghs_list.append(False)
                third_box_ghs_box_height_list.append(0.0)
                fourth_box_ghs_box_height_list.append(0.0)

                fifth_inner_box_height_list.append(22.0)
                sixth_inner_box_height_list.append(22.0)
                outer_box_height_list.append(60)

            group_box_height_list.append(outer_box_height_list[tmp_index] + 25)

            tmp_index += 1

        if fill_single_page:
            amount_on_page: typing.Dict[str, int] = {
                "DIN A4 (Portrait)": 3,
                "DIN A4 (Landscape)": 2,
                "Letter (Portrait)": 2,
                "Letter (Landscape)": 2
            }
            set_amount = int(amount_on_page[paper_format])
            body_height = 0
        else:
            set_amount = quantity
            body_height = 1

        object_amount = len(username_list)
        html = flask.render_template(
            "labels/mixed_format_labels.html",
            sample_code_list=sample_code_list,
            username_list=username_list,
            object_name_list=object_name_list,
            creation_date_list=creation_date_list,
            qr_code_uri_list=qr_code_uri_list,
            hazard_list=hazard_list,
            paper_width=paper_width,
            paper_height=paper_height,
            set_amount=set_amount,
            horizontal_label_margin=horizontal_label_margin,
            vertical_label_margin=vertical_label_margin,
            GHS_IMAGE_URIS=GHS_IMAGE_URIS,
            first_box_min_width_list=first_box_min_width_list,
            second_box_min_width_list=second_box_min_width_list,
            third_box_width=third_box_width,
            forth_box_width=forth_box_width,
            fifth_box_width=fifth_box_width,
            sixth_box_width=sixth_box_width,
            outer_box_width=outer_box_width,
            ghs_width=ghs_width,
            fifth_box_qrcode_box_width_list=fifth_box_qrcode_box_width_list,
            fifth_box_ghs_box_width_list=fifth_box_ghs_box_width_list,
            sixth_box_qrcode_box_width=sixth_box_qrcode_box_width,
            sixth_box_ghs_box_width=sixth_box_ghs_box_width,
            has_ghs_list=has_ghs_list,
            fifth_inner_box_height_list=fifth_inner_box_height_list,
            sixth_inner_box_height_list=sixth_inner_box_height_list,
            third_box_qrcode_box_height_list=third_box_qrcode_box_height_list,
            third_box_ghs_box_height_list=third_box_ghs_box_height_list,
            fourth_box_qrcode_box_height_list=fourth_box_qrcode_box_height_list,
            fourth_box_ghs_box_height_list=fourth_box_ghs_box_height_list,
            object_amount=object_amount,
            group_box_height_list=group_box_height_list,
            ghs_amount_list=ghs_amount_list,
            FONT_URIS=FONT_URIS,
            qrcode_width=qrcode_width,
            text_extra_width_list=text_extra_width_list,
            text_extra_width_list_qr=text_extra_width_list_qr,
            body_height=body_height,
        )

    elif create_only_qr_codes:
        object_id = list(object_specifications.keys())[0]
        object_name = object_specifications[object_id]["object_name"]
        object_url = object_specifications[object_id]["object_url"]
        qr_code_uri = []

        for quantity_index in range(1, quantity + 1):
            if custom_qr_code_texts and f"{object_id}_{quantity_index}" in custom_qr_code_texts:
                url = custom_qr_code_texts[f"{object_id}_{quantity_index}"]
            elif only_id_qr_code and add_label_number:
                url = f"{object_id} {quantity_index}"
                if add_maximum_label_number:
                    url += f"_{quantity}"
            else:
                url = f"{object_id if only_id_qr_code else object_url}"
            image = qrcode.make(url, border=1)
            image_stream = io.BytesIO()
            image.save(image_stream, format='png')
            image_stream.seek(0)
            qr_code_uri.append('data:image/png;base64,' + base64.b64encode(image_stream.read()).decode('utf-8'))

        box_height = max(math.floor(qr_code_width) + 1.225, 6)
        page_margin = [0.0, 0.0]

        if show_id_on_label:
            if add_label_number:
                if add_maximum_label_number:
                    box_width = qr_code_width + max((len(str(object_id)) * 2.8) + (2 * len(str(quantity))) + 6,
                                                    len(object_name) * 2.2)
                    if quantity >= 10:
                        box_width += 2.8
                    if quantity >= 100:
                        box_width += 2.8
                else:
                    box_width = qr_code_width + max(2.8 + (len(str(object_id)) * 2.8) + len(str(quantity)),
                                                    len(object_name) * 2.2)
                    if quantity >= 10:
                        box_width += 2.8
                    if quantity >= 100:
                        box_width += 2.8
        elif add_label_number:
            if add_maximum_label_number:
                box_width = qr_code_width + max((2 * len(str(quantity))) + 5.6, len(object_name) * 2.2)
                if quantity >= 100:
                    box_width += 2.8
                if quantity == 1000:
                    box_width += 2.8
            else:
                box_width = qr_code_width + max(len(str(quantity)) + 2.8, len(object_name) * 2.2)
                if quantity >= 100:
                    box_width += 2.8

        if label_dimension is not None:
            paper_format = PAGE_SIZE_KEYS[label_dimension["paper_format"]]
            paper_width, paper_height = PAGE_SIZES[paper_format]
            qr_code_width = label_dimension["qr_code_width"]
            box_width = label_dimension["label_width"]
            box_height = label_dimension["label_height"]
            horizontal_label_margin = label_dimension["margin_horizontal"]
            vertical_label_margin = label_dimension["margin_vertical"]
            labels_in_row = label_dimension["labels_in_row"]
            labels_in_col = label_dimension["labels_in_col"]
            page_margin = [paper_height - (labels_in_col * (box_height + vertical_label_margin)),
                           paper_width - (labels_in_row * (box_width + horizontal_label_margin))]

        qrcode_width = qr_code_width + 0.5
        qr_code_top = (box_height - qr_code_width) / 2
        quantity_digits = math.floor(math.log10(quantity)) + 1
        if label_dimension:
            html = flask.render_template(
                "labels/custom_dimension_qrcode_labels.html",
                box_width=box_width,
                qr_code_uri=qr_code_uri,
                object_id=object_id,
                box_height=box_height,
                paper_width=paper_width,
                paper_height=paper_height,
                object_name=object_name,
                vertical_label_margin=vertical_label_margin,
                horizontal_label_margin=horizontal_label_margin,
                qrcode_width=qrcode_width,
                num_labels=quantity,
                show_id=show_id_on_label,
                add_label_nr=add_label_number,
                add_maximum_label_nr=add_maximum_label_number,
                qr_code_top=qr_code_top,
                FONT_URIS=FONT_URIS,
                quantity_digits=quantity_digits,
                page_margin=page_margin,
            )
        else:
            html = flask.render_template(
                "labels/qrcode_labels.html",
                qr_code_uri=qr_code_uri,
                object_id=object_id,
                paper_width=paper_width,
                paper_height=paper_height,
                object_name=object_name,
                qrcode_width=qrcode_width,
                num_labels=quantity,
                show_id=show_id_on_label,
                add_label_nr=add_label_number,
                add_maximum_label_nr=add_maximum_label_number,
                qr_code_top=qr_code_top,
                FONT_URIS=FONT_URIS,
                quantity_digits=quantity_digits,
            )

    else:
        box_width = max(label_width, min_label_width) - 0.4
        ghs_width = qr_code_width / 2
        ghs_box_width = box_width
        boxqrcode_height = 18.0

        if centered:
            if ghs_classes_side_by_side:
                ghs_box_width = ghs_box_width / 2
        else:
            if ghs_classes_side_by_side:
                ghs_box_width = ghs_width * 2
            else:
                ghs_box_width = ghs_width * 2 - 0.5
                boxqrcode_height = 19.25

        tmp_index = 0
        for object_id in object_specifications:
            username_list.append(object_specifications[object_id]["creation_user"])
            object_name_list.append(object_specifications[object_id]["object_name"])
            creation_date_list.append(object_specifications[object_id]["creation_date"])
            object_url_list.append(object_specifications[object_id]["object_url"])
            hazard_list.append(object_specifications[object_id]["ghs_classes"])
            sample_code_list.append(object_id)

            url = object_url_list[tmp_index]
            image = qrcode.make(url, border=1)
            image_stream = io.BytesIO()
            image.save(image_stream, format='png')
            image_stream.seek(0)
            qr_code_uri_list.append('data:image/png;base64,' + base64.b64encode(image_stream.read()).decode('utf-8'))

            if centered:
                if ghs_classes_side_by_side:  # centered and ghs_classes_side_by_side
                    if len(hazard_list[tmp_index]) == 0:
                        ghs_box_width = 0
                else:  # centered and not ghs_classes_side_by_side
                    if len(hazard_list[tmp_index]) == 0:
                        boxqrcode_height += 0.1
            ghs_amount_list.append(len(hazard_list[tmp_index]))
            tmp_index += 1

        object_amount = len(username_list)

        if quantity == 1 and fill_single_page:
            num_labels = 100  # excess labels are clipped once the first page is full
            outer_box_width = paper_width - 7.5
        else:
            num_labels = quantity
            outer_box_width = paper_width - 10

        body_height = 0 if fill_single_page else 1

        html = flask.render_template(
            "labels/fixed_width_labels.html",
            box_width=box_width,
            object_amount=object_amount,
            min_label_height=min_label_height,
            qr_code_uri_list=qr_code_uri_list,
            paper_width=paper_width,
            paper_height=paper_height,
            hazard_list=hazard_list,
            GHS_IMAGE_URIS=GHS_IMAGE_URIS,
            ghs_width=ghs_width,
            sample_code_list=sample_code_list,
            username_list=username_list,
            object_name_list=object_name_list,
            creation_date_list=creation_date_list,
            ghs_amount_list=ghs_amount_list,
            ghs_classes_side_by_side=ghs_classes_side_by_side,
            centered=centered,
            ghs_box_width=ghs_box_width,
            num_labels=num_labels,
            outer_box_width=outer_box_width,
            FONT_URIS=FONT_URIS,
            boxqrcode_height=boxqrcode_height,
            body_height=body_height,
            qrcode_width=15
        )

    return typing.cast(bytes, HTML(
        string=html, base_url="img"
    ).write_pdf(presentational_hints=True))



# coding: utf-8
"""

"""
import copy
import json
import typing

import flask
import flask_login
import werkzeug
from flask_wtf import FlaskForm
from wtforms import BooleanField, StringField, SelectField, IntegerField, SelectMultipleField
from wtforms.validators import InputRequired, ValidationError, DataRequired
from flask_babel import _, force_locale

from . import frontend
from .permission_forms import handle_permission_forms, set_up_permissions_forms
from .. import models
from ..logic.action_permissions import get_action_permissions_for_all_users, get_user_action_permissions, set_action_permissions_for_all_users, get_action_permissions_for_groups, get_action_permissions_for_projects, get_action_permissions_for_users, get_sorted_actions_for_user
from ..logic.actions import Action, create_action, get_action, update_action
from ..logic.action_types import get_action_type, get_action_types
from ..logic.action_translations import get_action_translations_for_action, set_action_translation, delete_action_translation
from ..logic.languages import get_languages, get_language, Language
from ..logic.components import get_component
from ..logic.favorites import get_user_favorite_action_ids, get_user_favorite_instrument_ids
from ..logic.instruments import get_user_instruments, get_instrument
from ..logic.markdown_images import mark_referenced_markdown_images_as_permanent
from ..logic import errors, users
from ..logic.schemas.validate_schema import validate_schema
from ..logic.schemas.templates import reverse_substitute_templates, find_invalid_template_paths
from ..logic.settings import get_user_setting, set_user_settings, get_user_settings
from ..logic.topics import get_topics, get_topic, set_action_topics
from ..logic.users import get_users, get_user
from ..logic.groups import get_group
from ..logic.projects import get_project
from .users.forms import ToggleFavoriteActionForm
from .utils import check_current_user_is_not_readonly, get_groups_form_data, parse_filter_id_params, build_modified_url
from ..utils import FlaskResponseT
from ..logic.markdown_to_html import markdown_to_safe_html
from ..logic.utils import get_translated_text
from .. import logic, babel
from ..models import Permissions

__author__ = 'Florian Rhiem <f.rhiem@fz-juelich.de>'

ACTION_LIST_FILTER_PARAMETERS = (
    'action_list_filters',
    'topic_ids'
)


class ActionForm(FlaskForm):
    type = IntegerField()
    instrument = SelectField()
    topics = SelectMultipleField()
    schema = StringField(validators=[InputRequired()])
    is_public = BooleanField()
    is_user_specific = BooleanField(default=True)
    is_markdown = BooleanField(default=None)
    is_hidden = BooleanField(default=None)
    short_description_is_markdown = BooleanField(default=None)
    translations = StringField(validators=[DataRequired()])
    usable_by = SelectField(choices=['with_permissions', 'admins', 'nobody'])
    objects_readable_by_all_users_by_default = BooleanField(default=None)
    use_json_editor = BooleanField(default=None)
    use_instrument_topics = BooleanField(default=None)

    def validate_type(form, field: IntegerField) -> None:
        try:
            action_type_id = int(field.data)
        except ValueError:
            raise ValidationError(_("Invalid action type"))
        try:
            action_type = get_action_type(action_type_id)
        except errors.ActionTypeDoesNotExistError:
            raise ValidationError(_("Unknown action type"))
        if action_type.admin_only and not flask_login.current_user.is_admin:
            raise ValidationError(_("Actions with this type can only be created or edited by administrators."))


def _parse_action_list_filters(
        params: werkzeug.datastructures.MultiDict[str, str],
        valid_topic_ids: typing.List[int]
) -> typing.Tuple[
    bool,
    typing.Optional[typing.List[int]],
]:
    FALLBACK_RESULT = False, None
    success, filter_topic_ids = parse_filter_id_params(
        params=params,
        param_aliases=['topic_ids'],
        valid_ids=valid_topic_ids,
        id_map={},
        multi_params_error='',
        parse_error=_('Unable to parse topic IDs.'),
        invalid_id_error=_('Invalid topic ID.')
    )
    if not success:
        return FALLBACK_RESULT

    return (
        True,
        filter_topic_ids,
    )


@frontend.route('/actions/')
@flask_login.login_required
def actions() -> FlaskResponseT:
    action_type_id_str = flask.request.args.get('t', None)
    action_type = None

    user_is_instrument_scientist = bool(get_user_instruments(flask_login.current_user.id, exclude_hidden=True))

    action_type_id: typing.Optional[int]
    if action_type_id_str is not None:
        try:
            action_type_id = int(action_type_id_str)
        except ValueError:
            # ensure old links still function
            action_type_id = {
                'samples': models.ActionType.SAMPLE_CREATION,
                'measurements': models.ActionType.MEASUREMENT,
                'simulations': models.ActionType.SIMULATION
            }.get(action_type_id_str)
        if action_type_id is not None:
            try:
                action_type = get_action_type(
                    action_type_id=action_type_id
                )
            except errors.ActionTypeDoesNotExistError:
                # fall back to all objects
                action_type_id = None
    else:
        action_type_id = None
    user_id_str = flask.request.args.get('user_id', None)
    if user_id_str is not None:
        try:
            user_id = int(user_id_str)
        except ValueError:
            user_id = None
            flask.flash(_('Invalid user ID.'), 'error')
    else:
        user_id = None
    if user_id is not None:
        try:
            users.check_user_exists(user_id)
        except errors.UserDoesNotExistError:
            user_id = None
            flask.flash(_('Invalid user ID.'), 'error')
    actions = get_sorted_actions_for_user(
        user_id=flask_login.current_user.id,
        action_type_id=action_type_id,
        owner_id=user_id,
    )
    if action_type_id is None:
        # exclude actions of types that generally do not allow creating objects
        actions = [
            action
            for action in actions
            if action.schema is not None and action.type is not None and not action.type.disable_create_objects
        ]
    filter_usable_actions = 'can_create_objects' in flask.request.args
    if filter_usable_actions:
        # exclude actions that cannot be used to create objects by the current user
        actions = [
            action
            for action in actions
            if action.schema is not None and action.type is not None and not action.type.disable_create_objects and not action.disable_create_objects and not (action.admin_only and not flask_login.current_user.is_admin)
        ]

    topics = get_topics()
    valid_topic_ids = [topic.id for topic in topics]

    if 'action_list_filters' in flask.request.args or any(any(flask.request.args.getlist(param)) for param in ACTION_LIST_FILTER_PARAMETERS):
        (
            success,
            filter_topic_ids,
        ) = _parse_action_list_filters(
            params=flask.request.args,
            valid_topic_ids=valid_topic_ids
        )
        if not success:
            return flask.abort(400)
    else:
        filter_topic_ids = get_user_settings(user_id=flask_login.current_user.id)['DEFAULT_ACTION_LIST_FILTERS'].get('filter_topic_ids', [])

    if filter_topic_ids is None:
        filter_topic_ids = []

    if filter_topic_ids:
        actions = [
            action
            for action in actions
            if any(topic in [topic.id for topic in action.topics] for topic in filter_topic_ids)
        ]

    filter_topic_infos = []
    if filter_topic_ids:
        for topic_id in filter_topic_ids:
            topic = get_topic(topic_id)
            filter_topic_infos.append({
                'name': get_translated_text(topic.name, default=_('Unnamed Topic')),
                'url': flask.url_for('.topic', topic_id=topic_id)
            })

    user_favorite_action_ids = get_user_favorite_action_ids(flask_login.current_user.id)
    toggle_favorite_action_form = ToggleFavoriteActionForm()
    return flask.render_template(
        'actions/actions.html',
        actions=actions,
        action_type=action_type,
        action_permissions=action_permissions,
        user_is_instrument_scientist=user_is_instrument_scientist,
        Permissions=Permissions,
        user_favorite_action_ids=user_favorite_action_ids,
        toggle_favorite_action_form=toggle_favorite_action_form,
        get_user=get_user,
        get_component=get_component,
        topics=topics,
        filter_topic_infos=filter_topic_infos,
        filter_topic_ids=filter_topic_ids
    )


@frontend.route('/actions/', methods=['POST'])
@flask_login.login_required
def save_action_list_defaults() -> FlaskResponseT:
    if 'save_default_action_filters' in flask.request.form:
        topics = get_topics()
        valid_topic_ids = [topic.id for topic in topics]
        success, filter_topic_ids = _parse_action_list_filters(
            params=flask.request.form,
            valid_topic_ids=valid_topic_ids
        )
        if not success:
            return flask.abort(400)
        set_user_settings(
            user_id=flask_login.current_user.id,
            data={
                'DEFAULT_ACTION_LIST_FILTERS': {
                    'filter_topic_ids': filter_topic_ids,
                }
            }
        )
        return flask.redirect(build_modified_url('.actions', blocked_parameters=ACTION_LIST_FILTER_PARAMETERS))
    if 'clear_default_action_filters' in flask.request.form:
        set_user_settings(
            user_id=flask_login.current_user.id,
            data={
                'DEFAULT_ACTION_LIST_FILTERS': {}
            }
        )
        return flask.redirect(build_modified_url('.actions', blocked_parameters=ACTION_LIST_FILTER_PARAMETERS))
    return flask.abort(400)


@frontend.route('/actions/<int:action_id>', methods=['GET', 'POST'])
@flask_login.login_required
def action(action_id: int) -> FlaskResponseT:
    try:
        action = get_action(action_id)
    except errors.ActionDoesNotExistError:
        return flask.abort(404)
    permissions = get_user_action_permissions(action_id, flask_login.current_user.id)
    if Permissions.READ not in permissions:
        return flask.abort(403)
    may_edit = Permissions.WRITE in permissions and action.fed_id is None

    if action.type is not None and action.type.admin_only and not flask_login.current_user.is_admin:
        may_edit = False
    may_grant = Permissions.GRANT in permissions
    mode = flask.request.args.get('mode', None)
    if mode == 'edit':
        action_schema = action.schema
        if action_schema:
            original_schema = copy.deepcopy(action_schema)
            try:
                reverse_substitute_templates(action_schema)
            except errors.ActionDoesNotExistError:
                action_schema = original_schema
                flask.flash(_('The used template does not exist anymore. Use the JSON editor to edit the existing action.'), 'error')
                if get_user_setting(flask_login.current_user.id, "USE_SCHEMA_EDITOR"):
                    flask.abort(400)
        check_current_user_is_not_readonly()
        if not may_edit:
            if action.fed_id is not None:
                flask.flash(_('Editing imported actions is not yet supported.'), 'error')
            return flask.abort(403)
        return show_action_form(action, action_schema=action_schema)

    action_is_favorite = action.id in get_user_favorite_action_ids(flask_login.current_user.id)
    return flask.render_template(
        'actions/action.html',
        action=action,
        action_is_favorite=action_is_favorite,
        may_edit=may_edit,
        may_grant=may_grant,
        is_public=Permissions.READ in get_action_permissions_for_all_users(action_id),
        get_user=get_user,
        get_component=get_component
    )


@frontend.route('/actions/new/', methods=['GET', 'POST'])
@flask_login.login_required
def new_action() -> FlaskResponseT:
    check_current_user_is_not_readonly()
    action_type_id_str = flask.request.args.get('action_type_id')
    if action_type_id_str is not None:
        try:
            action_type_id = int(action_type_id_str)
        except ValueError:
            action_type_id = None
    else:
        action_type_id = None
    topic_ids_str = flask.request.args.get('topic_ids')
    if topic_ids_str is not None:
        try:
            topic_ids = [int(topic_id_str) for topic_id_str in topic_ids_str.split(',')]
        except ValueError:
            topic_ids = None
    else:
        topic_ids = None
    instrument_id_str = flask.request.args.get('instrument_id', None)
    if instrument_id_str is not None:
        try:
            instrument_id = int(instrument_id_str)
        except ValueError:
            instrument_id = None
    else:
        instrument_id = None
    if instrument_id:
        try:
            user_instrument_ids = get_user_instruments(flask_login.current_user.id, exclude_hidden=True)
            if instrument_id in user_instrument_ids:
                instrument = get_instrument(instrument_id)
                if instrument.component_id is not None:
                    flask.flash(_('Using imported instruments to create actions is not supported.'), 'error')
                    instrument_id = None
            else:
                flask.flash(_('Insufficient permissions to use the requested instrument.'), 'error')
                instrument_id = None
        except errors.InstrumentDoesNotExistError:
            flask.flash(_('The requested instrument does not exist.'), 'error')
    previous_action = None
    previous_action_id_str = flask.request.args.get('previous_action_id', None)
    if previous_action_id_str is not None:
        try:
            previous_action_id = int(previous_action_id_str)
        except ValueError:
            previous_action_id = None
    else:
        previous_action_id = None
    if previous_action_id:
        try:
            permissions = get_user_action_permissions(previous_action_id, flask_login.current_user.id)
            if Permissions.READ in permissions:
                previous_action = get_action(previous_action_id)
            else:
                flask.flash(_('Insufficient permissions to use the requested action as a template.'), 'error')
                previous_action = None
        except errors.ActionDoesNotExistError:
            flask.flash(_('The requested action does not exist.'), 'error')
        else:
            if previous_action is not None:
                if previous_action.type is None or previous_action.schema is None:
                    previous_action = None
                elif previous_action.type.admin_only and not flask_login.current_user.is_admin:
                    flask.flash(_('Only administrators can create actions of this type.'), 'error')
                    return flask.redirect(flask.url_for('.action', action_id=previous_action_id))
    return show_action_form(None, previous_action, action_type_id, topic_ids=topic_ids, instrument_id=instrument_id)


def _get_lines_for_path(schema: typing.Dict[str, typing.Any], path: typing.List[str]) -> typing.Optional[typing.Set[int]]:
    schema_entry = schema
    parent = None
    key = None
    for property in path:
        if property == '[?]' and 'items' in schema_entry and isinstance(schema_entry['items'], dict):
            parent = schema_entry
            key = 'items'
            if parent is not None and key in parent:
                schema_entry = parent[key]
            else:
                return None
        elif 'properties' in schema_entry and property in schema_entry['properties'] and isinstance(schema_entry['properties'][property], dict):
            parent = schema_entry['properties']
            key = property
            if parent is not None and key in parent:
                schema_entry = parent[key]
            else:
                return None
        else:
            break
    if parent is None or key is None:
        return None

    parent[key] = {"INTERNAL_MARKER_START": None, "INTERNAL_MARKER_CONTENT": parent[key], "INTERNAL_MARKER_END": None}

    schema_json = json.dumps(schema, indent=2)
    in_error = False
    error_lines = []
    skip_lines = []
    schema_json_lines = schema_json.splitlines(keepends=False)
    for i, line in enumerate(schema_json_lines):
        if line.endswith('"INTERNAL_MARKER_START": null,'):
            skip_lines.append(i)
            skip_lines.append(i + 1)
            in_error = True
        if in_error:
            error_lines.append(i)
        if line.endswith('"INTERNAL_MARKER_END": null'):
            skip_lines.append(i - 1)
            skip_lines.append(i)
            in_error = False

    new_error_lines: typing.Set[int] = set()
    for i in reversed(error_lines):
        if i in skip_lines:
            new_error_lines = {i - 1 for i in new_error_lines}
        else:
            new_error_lines.add(i + 1)
            new_error_lines.add(i)
            new_error_lines.add(i - 1)
    return new_error_lines


def show_action_form(
        action: typing.Optional[Action] = None,
        previous_action: typing.Optional[Action] = None,
        action_type_id: typing.Optional[int] = None,
        action_schema: typing.Optional[typing.Dict[str, typing.Any]] = None,
        topic_ids: typing.Optional[typing.List[int]] = None,
        instrument_id: typing.Optional[int] = None,
) -> FlaskResponseT:
    action_translations = []
    load_translations = False

    languages = get_languages(only_enabled_for_input=True)
    allowed_language_ids = [
        language.id
        for language in languages
    ]
    tags_translations = {}
    hazards_translations = {}
    translation_lang_codes = [locale.language for locale in babel.list_translations()]
    for language in languages:
        if language.lang_code in translation_lang_codes:
            with force_locale(language.lang_code):
                tags_translations[language.lang_code] = _('Tags')
                hazards_translations[language.lang_code] = _('GHS Hazards')

    topics = get_topics()

    if action is not None:
        action_translations = get_action_translations_for_action(action.id, use_fallback=True)
        load_translations = True
        schema_json = json.dumps(action_schema, indent=2)
        submit_text = "Save"
    elif previous_action is not None:
        action_translations = get_action_translations_for_action(previous_action.id, use_fallback=True)
        load_translations = True
        schema_json = json.dumps(previous_action.schema, indent=2)
        submit_text = "Create"
    else:
        schema_json = json.dumps({
            'title': {
                'en': 'Object Information'
            },
            'type': 'object',
            'properties': {
                'name': {
                    'title': {
                        'en': 'Name'
                    },
                    'type': 'text'
                }
            },
            'required': ['name'],
            'propertyOrder': [
                'name'
            ]
        }, indent=2)
        submit_text = "Create"
    may_set_user_specific = action is None and flask_login.current_user.is_admin
    schema = None
    error_message = None
    action_form = ActionForm()
    instrument_is_fed = {}
    instrument_is_fav = {}

    sample_action_type = get_action_type(
        action_type_id=models.ActionType.SAMPLE_CREATION
    )
    measurement_action_type = get_action_type(
        action_type_id=models.ActionType.MEASUREMENT
    )
    english = get_language(Language.ENGLISH)
    action_language_ids = [translation.language_id for translation in action_translations]
    action_translations_by_id = {
        action_translation.language_id: action_translation
        for action_translation in action_translations
    }
    user_favorite_instrument_ids = get_user_favorite_instrument_ids(flask_login.current_user.id)
    action_form.topics.choices = [
        (str(topic.id), get_translated_text(topic.name))
        for topic in topics
    ]
    if not action_form.is_submitted():
        action_form.use_json_editor.data = get_user_setting(flask_login.current_user.id, "USE_SCHEMA_EDITOR")
        action_form.topics.data = []
    if action is not None:
        if action.instrument is not None:
            action_form.instrument.choices = [
                (str(action.instrument_id), get_translated_text(action.instrument.name, default=_('Unnamed Instrument')))
            ]
            instrument_is_fed[str(action.instrument_id)] = action.instrument.component_id is not None
            instrument_is_fav[str(action.instrument_id)] = action.instrument_id in user_favorite_instrument_ids
            action_form.instrument.data = str(action.instrument_id)
        else:
            action_form.instrument.choices = [('-1', '-')]
            action_form.instrument.data = str(-1)
        action_form.type.data = action.type_id
    else:
        if flask.current_app.config['DISABLE_INSTRUMENTS']:
            action_form.instrument.choices = [('-1', '-')]
            action_form.instrument.data = '-1'
        else:
            user_instrument_ids = get_user_instruments(flask_login.current_user.id, exclude_hidden=True)
            for user_instrument_id in user_instrument_ids:
                instrument_is_fed[str(user_instrument_id)] = get_instrument(user_instrument_id).component_id is not None
                instrument_is_fav[str(user_instrument_id)] = user_instrument_id in user_favorite_instrument_ids
            user_instruments = [
                (str(instrument_id), get_translated_text(get_instrument(instrument_id).name, default=_('Unnamed Instrument')))
                for instrument_id in user_instrument_ids
            ]
            user_instruments.sort(key=lambda instrument_data: (
                not instrument_is_fav[instrument_data[0]],
                instrument_data[1]
            ))
            action_form.instrument.choices = [('-1', '-')] + user_instruments
            if action_form.instrument.data is None or action_form.instrument.data == str(None):
                if previous_action is not None and previous_action.instrument_id in user_instrument_ids:
                    action_form.instrument.data = str(previous_action.instrument_id)
                elif instrument_id is not None:
                    action_form.instrument.data = str(instrument_id)
                else:
                    action_form.instrument.data = '-1'
    form_is_valid = False
    if not action_form.is_submitted():
        action_form.type.data = action_type_id
        if topic_ids and not flask.current_app.config['DISABLE_TOPICS']:
            action_form.topics.data = [str(topic_id) for topic_id in topic_ids]
        else:
            action_form.topics.data = []
    elif action_form.validate_on_submit():
        form_is_valid = True
        action_type_id = action_form.type.data
    else:
        action_type_id = None
    if action_type_id is None:
        form_is_valid = False

    if not action_form.is_submitted():
        if action is not None:
            action_form.is_hidden.data = action.is_hidden
            action_form.is_markdown.data = action.description_is_markdown
            action_form.short_description_is_markdown.data = action.short_description_is_markdown
            action_form.type.data = action.type_id
            action_form.is_public.data = Permissions.READ in get_action_permissions_for_all_users(action.id)
            if action.disable_create_objects:
                action_form.usable_by.data = 'nobody'
            elif action.admin_only:
                action_form.usable_by.data = 'admins'
            else:
                action_form.usable_by.data = 'with_permissions'
            action_form.objects_readable_by_all_users_by_default.data = action.objects_readable_by_all_users_by_default
            action_form.topics.data = [str(topic.id) for topic in action.topics]
            action_form.use_instrument_topics.data = action.use_instrument_topics
        elif previous_action is not None:
            action_form.is_hidden.data = False
            action_form.is_markdown.data = previous_action.description_is_markdown
            action_form.short_description_is_markdown.data = previous_action.short_description_is_markdown
            action_form.type.data = previous_action.type_id
            action_form.is_public.data = Permissions.READ in get_action_permissions_for_all_users(previous_action.id)
            action_form.usable_by.data = 'with_permissions'
            action_form.objects_readable_by_all_users_by_default.data = previous_action.objects_readable_by_all_users_by_default
            action_form.topics.data = [str(topic.id) for topic in previous_action.topics]
            action_form.use_instrument_topics.data = False
            if not action_form.topics.data and previous_action.topics:
                action_form.topics.data = [str(topic.id) for topic in previous_action.topics]
        if flask.current_app.config['DISABLE_TOPICS']:
            action_form.topics.choices = []
            action_form.topics.data = []

    if action_form.schema.data:
        schema_json = action_form.schema.data
    else:
        action_form.schema.data = schema_json
    error_lines = None
    if schema_json:
        all_lines = set(range(1, 1 + len(schema_json.splitlines())))
        error_lines = set()
        if 'INTERNAL_MARKER' in schema_json:
            error_lines = all_lines
            error_message = "Failed to parse schema: Must not contain INTERNAL_MARKER"
        else:
            try:
                schema = json.loads(schema_json)
            except json.JSONDecodeError as e:
                error_lines = {e.lineno}
                error_message = f"Failed to parse as JSON: {e.msg}"
            except Exception as e:
                error_lines = all_lines
                error_message = f"Failed to parse as JSON: {str(e)}"
        if schema is not None:
            try:
                invalid_template_paths = find_invalid_template_paths(schema, flask_login.current_user.id)
                if invalid_template_paths:
                    raise errors.ValidationError('insufficient permissions for template action', invalid_template_paths[0])
                validate_schema(schema, invalid_template_action_ids=[] if action is None else [action.id], strict=True)
            except errors.ValidationError as e:
                error_message = e.message
                if not e.paths:
                    error_lines = all_lines
                else:
                    schema_json = json.dumps(schema, indent=2)
                    action_form.schema.data = schema_json
                    for path in e.paths:
                        new_error_lines = _get_lines_for_path(schema, path)
                        if new_error_lines is None:
                            error_lines = all_lines
                            break
                        else:
                            error_lines.update(new_error_lines)
                    else:
                        error_lines = {i + 1 for i in error_lines}
            except Exception as e:
                error_lines = all_lines
                error_message = _("Unknown error: ") + str(e)
    if schema is not None and error_message is None and form_is_valid:
        # First block for validation
        try:
            translation_data = json.loads(action_form.translations.data)
        except Exception:
            flask.flash(_("Something went wrong"), 'error')
            return flask.render_template(
                'actions/action_form.html',
                current_action=action,
                action_form=action_form,
                action_translations=action_translations_by_id,
                action_language_ids=action_language_ids,
                action_types=get_action_types(),
                error_message=error_message,
                schema_json=schema_json,
                submit_text=submit_text,
                sample_action_type=sample_action_type,
                measurement_action_type=measurement_action_type,
                may_change_type=action is None,
                may_change_instrument=action is None,
                may_set_user_specific=may_set_user_specific,
                languages=languages,
                tags_translations=tags_translations,
                hazards_translations=hazards_translations,
                load_translations=load_translations,
                ENGLISH=english,
                instrument_is_fed=instrument_is_fed,
                instrument_is_fav=instrument_is_fav,
                error_lines=list(error_lines or []),
            )
        else:
            translation_keys = {'language_id', 'name', 'description', 'short_description'}

            if not isinstance(translation_data, list):
                translation_data = ()

            for translation in translation_data:
                if not isinstance(translation, dict):
                    continue
                if set(translation.keys()) != translation_keys:
                    continue

                language_id = int(translation['language_id'])
                name = translation['name'].strip()
                description = translation['description'].strip()
                short_description = translation['short_description'].strip()

                if language_id == Language.ENGLISH:
                    error = False
                    if not isinstance(name, str) or not isinstance(description, str) \
                            or not isinstance(short_description, str):
                        flask.flash(_('Please input a name.'))
                        error = True
                    if len(name) < 1 or len(name) > 100:
                        error = True
                        flask.flash(_('Please input a name with a length between 1 and 100.'))
                    if error:
                        return flask.render_template(
                            'actions/action_form.html',
                            current_action=action,
                            action_form=action_form,
                            action_translations=action_translations_by_id,
                            action_language_ids=action_language_ids,
                            action_types=get_action_types(),
                            error_message=error_message,
                            schema_json=schema_json,
                            submit_text=submit_text,
                            sample_action_type=sample_action_type,
                            measurement_action_type=measurement_action_type,
                            may_change_type=action is None,
                            may_change_instrument=action is None,
                            may_set_user_specific=may_set_user_specific,
                            languages=languages,
                            tags_translations=tags_translations,
                            hazards_translations=hazards_translations,
                            load_translations=load_translations,
                            ENGLISH=english,
                            instrument_is_fed=instrument_is_fed,
                            instrument_is_fav=instrument_is_fav,
                            error_lines=list(error_lines or []),
                        )
        instrument_id_str = action_form.instrument.data
        topic_id_strs = action_form.topics.data
        is_public = action_form.is_public.data
        is_hidden = action_form.is_hidden.data
        is_markdown = action_form.is_markdown.data
        short_description_is_markdown = action_form.short_description_is_markdown.data
        admin_only = action_form.usable_by.data == 'admins'
        disable_create_objects = action_form.usable_by.data == 'nobody'
        objects_readable_by_all_users_by_default = action_form.objects_readable_by_all_users_by_default.data
        use_instrument_topics = action_form.use_instrument_topics.data

        if instrument_id_str is not None:
            try:
                instrument_id = int(instrument_id_str)
            except ValueError:
                instrument_id = None
            if instrument_id is not None and instrument_id < 0:
                instrument_id = None
        else:
            instrument_id = None

        try:
            topic_ids = [int(topic_id_str) for topic_id_str in topic_id_strs]
        except ValueError:
            topic_ids = []

        assert action_type_id is not None  # mypy does not infer this, but it is guaranteed by the code above
        if action is None:
            if action_form.is_user_specific.data or not may_set_user_specific:
                user_id = flask_login.current_user.id
            else:
                user_id = None
            action = create_action(
                action_type_id=action_type_id,
                schema=schema,
                instrument_id=instrument_id,
                user_id=user_id,
                is_hidden=is_hidden,
                description_is_markdown=is_markdown,
                short_description_is_markdown=short_description_is_markdown,
                admin_only=admin_only,
                disable_create_objects=disable_create_objects,
                objects_readable_by_all_users_by_default=objects_readable_by_all_users_by_default,
                use_instrument_topics=use_instrument_topics if instrument_id is not None else False
            )
        else:
            update_action(
                action_id=action.id,
                schema=schema,
                is_hidden=is_hidden,
                description_is_markdown=is_markdown,
                short_description_is_markdown=short_description_is_markdown,
                admin_only=admin_only,
                disable_create_objects=disable_create_objects,
                objects_readable_by_all_users_by_default=objects_readable_by_all_users_by_default,
                use_instrument_topics=use_instrument_topics if action.instrument_id is not None else False
            )
        if action.instrument_id is None or not use_instrument_topics:
            set_action_topics(action.id, topic_ids)
        set_action_permissions_for_all_users(action.id, Permissions.READ if is_public else Permissions.NONE)

        # After validation
        valid_translations = set()
        for translation in translation_data:
            if not isinstance(translation, dict):
                continue
            if set(translation.keys()) != translation_keys:
                continue

            language_id = int(translation['language_id'])
            name = translation['name'].strip()
            description = translation['description'].strip()
            short_description = translation['short_description'].strip()

            if language_id not in allowed_language_ids:
                continue

            if is_markdown:
                description_as_html = markdown_to_safe_html(description, anchor_prefix="instrument-description")
                mark_referenced_markdown_images_as_permanent(description_as_html)
            if short_description_is_markdown:
                short_description_as_html = markdown_to_safe_html(short_description,
                                                                  anchor_prefix="instrument-short-description")
                mark_referenced_markdown_images_as_permanent(short_description_as_html)

            new_translation = set_action_translation(
                language_id=language_id,
                action_id=action.id,
                name=name,
                description=description,
                short_description=short_description)
            valid_translations.add((new_translation.language_id, new_translation.action_id))

        for existing_translations in get_action_translations_for_action(action.id):
            if (existing_translations.language_id, existing_translations.action_id) not in valid_translations:
                delete_action_translation(
                    language_id=existing_translations.language_id,
                    action_id=existing_translations.action_id
                )

        flask.flash(_('The action was updated successfully.'), 'success')
        return flask.redirect(flask.url_for('.action', action_id=action.id))
    if action_form.translations.name in flask.request.form:
        # load translations from form
        try:
            translation_data = json.loads(action_form.translations.data)
            for translation in translation_data:
                translation['language_id'] = int(translation['language_id'])
                translation['language'] = get_language(translation['language_id'])
                action_translations_by_id[translation['language_id']] = translation
                if translation['language_id'] not in action_translations_by_id:
                    action_language_ids.append(translation['language_id'])
            if action_language_ids:
                load_translations = True
        except Exception:
            pass
    return flask.render_template(
        'actions/action_form.html',
        current_action=action,
        action_form=action_form,
        action_translations=action_translations_by_id,
        action_language_ids=action_language_ids,
        action_types=get_action_types(),
        error_message=error_message,
        schema_json=schema_json,
        submit_text=submit_text,
        sample_action_type=sample_action_type,
        measurement_action_type=measurement_action_type,
        may_change_type=action is None,
        may_change_instrument=action is None,
        may_set_user_specific=may_set_user_specific,
        languages=languages,
        tags_translations=tags_translations,
        hazards_translations=hazards_translations,
        load_translations=load_translations,
        ENGLISH=english,
        instrument_is_fed=instrument_is_fed,
        instrument_is_fav=instrument_is_fav,
        error_lines=list(error_lines or []),
    )


@frontend.route('/actions/<int:action_id>/permissions', methods=['GET', 'POST'])
@flask_login.login_required
def action_permissions(action_id: int) -> FlaskResponseT:
    try:
        action = get_action(action_id)
    except errors.ActionDoesNotExistError:
        return flask.abort(404)
    permissions = get_user_action_permissions(action_id, flask_login.current_user.id)
    if Permissions.READ not in permissions:
        return flask.abort(403)
    user_may_edit = Permissions.GRANT in permissions
    all_user_permissions = get_action_permissions_for_all_users(action_id=action.id)
    anonymous_user_permissions = Permissions.NONE
    user_permissions = get_action_permissions_for_users(
        action_id=action.id
    )
    if action.user_id is not None:
        # action owner has implicit GRANT permissions
        action_owner = get_user(action.user_id)
        if action_owner.is_readonly:
            user_permissions[action_owner.id] = Permissions.READ
        else:
            user_permissions[action_owner.id] = Permissions.GRANT
    group_permissions = get_action_permissions_for_groups(
        action_id=action.id
    )
    project_permissions = get_action_permissions_for_projects(
        action_id=action.id
    )
    if user_may_edit:
        (
            add_user_permissions_form,
            add_group_permissions_form,
            add_project_permissions_form,
            permissions_form
        ) = set_up_permissions_forms(
            resource_permissions=logic.action_permissions.action_permissions,
            resource_id=action_id,
            existing_all_user_permissions=all_user_permissions,
            existing_anonymous_user_permissions=anonymous_user_permissions,
            existing_user_permissions=user_permissions,
            existing_group_permissions=group_permissions,
            existing_project_permissions=project_permissions
        )

        users = get_users(exclude_hidden=not flask_login.current_user.is_admin or not flask_login.current_user.settings['SHOW_HIDDEN_USERS_AS_ADMIN'], exclude_fed=True, exclude_eln_import=True)
        users = [user for user in users if user.id not in user_permissions]
        users.sort(key=lambda user: user.id)

        show_groups_form, groups_treepicker_info = get_groups_form_data(
            basic_group_filter=lambda group: group.id not in group_permissions
        )

        show_projects_form, projects_treepicker_info = get_groups_form_data(
            project_group_filter=lambda group: group.id not in project_permissions
        )
        if flask.request.method.lower() == 'post':
            if handle_permission_forms(
                    logic.action_permissions.action_permissions,
                    action_id,
                    add_user_permissions_form,
                    add_group_permissions_form,
                    add_project_permissions_form,
                    permissions_form,
                    user_permissions,
                    group_permissions,
                    project_permissions,
            ):
                flask.flash(_('Successfully updated action permissions.'), 'success')
            else:
                flask.flash(_('Failed to update action permissions.'), 'error')
            return flask.redirect(flask.url_for('.action_permissions', action_id=action_id))
    else:
        permissions_form = None
        users = None
        add_user_permissions_form = None
        add_group_permissions_form = None
        add_project_permissions_form = None
        show_groups_form = False
        groups_treepicker_info = None
        show_projects_form = False
        projects_treepicker_info = None
        if flask.request.method.lower() == 'post':
            flask.flash(_('You need GRANT permissions to edit the permissions for this action.'), 'error')
            return flask.redirect(flask.url_for('.action_permissions', action_id=action_id))

    return flask.render_template(
        'actions/action_permissions.html',
        user_may_edit=user_may_edit,
        action=action,
        instrument=action.instrument,
        Permissions=Permissions,
        project_permissions=project_permissions,
        group_permissions=group_permissions,
        user_permissions=user_permissions,
        all_user_permissions=all_user_permissions,
        permissions_form=permissions_form,
        users=users,
        add_user_permissions_form=add_user_permissions_form,
        add_group_permissions_form=add_group_permissions_form,
        add_project_permissions_form=add_project_permissions_form,
        show_groups_form=show_groups_form,
        groups_treepicker_info=groups_treepicker_info,
        show_projects_form=show_projects_form,
        projects_treepicker_info=projects_treepicker_info,
        get_user=get_user,
        get_group=get_group,
        get_project=get_project,
    )



# coding: utf-8
"""
Add the show_in_object_filters column to the action_types table.
"""

import flask_sqlalchemy

from .utils import table_has_column


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if table_has_column('action_types', 'show_in_object_filters'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE action_types
        ADD show_in_object_filters BOOLEAN DEFAULT TRUE NOT NULL
    """))
    # Hide non-object-creating action types per default
    db.session.execute(db.text("""
        UPDATE action_types
        SET show_in_object_filters = FALSE
        WHERE disable_create_objects
    """))
    return True
